import inspect
import logging
import numpy as np
from typing import Any, List, Iterable, Optional, Tuple, Union, cast
import pandas as pd
from tqdm import tqdm
import os
from jaqpotpy.cfg import config

logger = logging.getLogger(__name__)
_print_threshold = 10


class Featurizer(object):
  """Abstract class for calculating a set of features for a datapoint.
  This class is abstract and cannot be invoked directly. You'll
  likely only interact with this class if you're a developer. In
  that case, you might want to make a child class which
  implements the `_featurize` method for calculating features for
  a single datapoints if you'd like to make a featurizer for a
  new datatype.
    References
  ----------
  .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
     O'Reilly Media (2019):.
  .. [2] https://github.com/deepchem/deepchem
  """

  def featurize(self,
                datapoints: Iterable[Any],
                log_every_n: int = 1000,
                **kwargs) -> np.ndarray:
    """Calculate features for datapoints.
    Parameters
    ----------
    datapoints: Iterable[Any]
      A sequence of objects that you'd like to featurize. Subclassses of
      `Featurizer` should instantiate the `_featurize` method that featurizes
      objects in the sequence.
    log_every_n: int, default 1000
      Logs featurization progress every `log_every_n` steps.
    Returns
    -------
    np.ndarray
      A numpy array containing a featurized representation of `datapoints`.
    """
    datapoints = list(datapoints)
    features = []
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, point in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):

      try:
        features.append(self._featurize(point, **kwargs))
      except:
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %d. Appending empty array")
        features.append(np.array([]))

    return np.asarray(features)

  def __call__(self, datapoints: Iterable[Any], **kwargs):
    """Calculate features for datapoints.
    `**kwargs` will get passed directly to `Featurizer.featurize`
    Parameters
    ----------
    datapoints: Iterable[Any]
      Any blob of data you like. Subclasss should instantiate this.
    """
    return self.featurize(datapoints, **kwargs)

  def _featurize(self, datapoint: Any, **kwargs):
    """Calculate features for a single datapoint.
    Parameters
    ----------
    datapoint: Any
      Any blob of data you like. Subclass should instantiate this.
    """
    raise NotImplementedError('Featurizer is not defined.')

  def _featurize_dataframe(self, datapoint: Any, **kwargs):
    """Calculate features for a single datapoint.
    Parameters
    ----------
    datapoint: Any
      Any blob of data you like. Subclass should instantiate this.
    """
    raise NotImplementedError('Featurizer is not defined.')

  def __repr__(self) -> str:
    """Convert self to repr representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
    CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
    >>> jt.descriptors.CGCNNFeaturizer()
    CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_info = ''
    for arg_name in args_names:
      value = self.__dict__[arg_name]
      # for str
      if isinstance(value, str):
        value = "'" + value + "'"
      # for list
      if isinstance(value, list):
        threshold = get_print_threshold()
        value = np.array2string(np.array(value), threshold=threshold)
      args_info += arg_name + '=' + str(value) + ', '
    return self.__class__.__name__ + '[' + args_info[:-2] + ']'

  def __str__(self) -> str:
    """Convert self to str representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
    CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
    >>> jt.descriptors.CGCNNFeaturizer()
    CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_num = len(args_names)
    args_default_values = [None for _ in range(args_num)]
    if args_spec.defaults is not None:
      defaults = list(args_spec.defaults)
      args_default_values[-len(defaults):] = defaults

    override_args_info = ''
    for arg_name, default in zip(args_names, args_default_values):
      if arg_name in self.__dict__:
        arg_value = self.__dict__[arg_name]
        # validation
        # skip list
        if isinstance(arg_value, list):
          continue
        if isinstance(arg_value, str):
          # skip path string
          if "\\/." in arg_value or "/" in arg_value or '.' in arg_value:
            continue
        # main logic
        if default != arg_value:
          override_args_info += '_' + arg_name + '_' + str(arg_value)
    return self.__class__.__name__ + override_args_info

class ComplexFeaturizer(Featurizer):
  """"
  Abstract class for calculating features for mol/protein complexes.
  """

  def featurize(self,
                datapoints: Optional[Iterable[Tuple[str, str]]] = None,
                log_every_n: int = 100,
                **kwargs) -> np.ndarray:
    """
    Calculate features for mol/protein complexes.
    Parameters
    ----------
    datapoints: Iterable[Tuple[str, str]]
      List of filenames (PDB, SDF, etc.) for ligand molecules and proteins.
      Each element should be a tuple of the form (ligand_filename,
      protein_filename).
    Returns
    -------
    features: np.ndarray
      Array of features
    """

    if 'complexes' in kwargs:
      datapoints = kwargs.get("complexes")
      raise DeprecationWarning(
          'Complexes is being phased out as a parameter, please pass "datapoints" instead.'
      )
    if not isinstance(datapoints, Iterable):
      datapoints = [cast(Tuple[str, str], datapoints)]
    features, failures, successes = [], [], []
    for idx, point in enumerate(datapoints):
      if idx % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % idx)
      try:
        features.append(self._featurize(point, **kwargs))
        successes.append(idx)
      except:
        logger.warning(
            "Failed to featurize datapoint %i. Appending empty array." % idx)
        features.append(np.zeros(1))
        failures.append(idx)

    # Find a successful featurization
    try:
      i = np.argmax([f.shape[0] for f in features])
      dtype = features[i].dtype
      shape = features[i].shape
      dummy_array = np.zeros(shape, dtype=dtype)
    except AttributeError:
      dummy_array = features[successes[0]]

    # Replace failed featurizations with appropriate array
    for idx in failures:
      features[idx] = dummy_array

    return np.asarray(features)

  def _featurize(self, datapoint: Optional[Tuple[str, str]] = None, **kwargs):
    """
    Calculate features for single mol/protein complex.
    Parameters
    ----------
    complex: Tuple[str, str]
      Filenames for molecule and protein.
    """
    raise NotImplementedError('Featurizer is not defined.')


class DataframeFeaturizer(object):
  """Abstract class for calculating a set of features for a datapoint.
  This class is abstract and cannot be invoked directly. You'll
  likely only interact with this class if you're a developer. In
  that case, you might want to make a child class which
  implements the `_featurize` method for calculating features for
  a single datapoints if you'd like to make a featurizer for a
  new datatype.
    References
  ----------
  .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
     O'Reilly Media (2019):.
  .. [2] https://github.com/deepchem/deepchem
  """

  def featurize(self,
                datapoints: Iterable[Any],
                log_every_n: int = 1000,
                **kwargs) -> Any:
    """Calculate features for datapoints.
    Parameters
    ----------
    datapoints: Iterable[Any]
      A sequence of objects that you'd like to featurize. Subclassses of
      `Featurizer` should instantiate the `_featurize` method that featurizes
      objects in the sequence.
    log_every_n: int, default 1000
      Logs featurization progress every `log_every_n` steps.
    Returns
    -------
    np.ndarray
      A numpy array containing a featurized representation of `datapoints`.
    """
    datapoints = list(datapoints)
    features = []
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, point in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)
      try:
        features.append(self._featurize(point, **kwargs))
      except:
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %d. Appending empty array")
        features.append(np.array([]))

    return np.asarray(features)

  def __call__(self, datapoints: Iterable[Any], **kwargs):
    """Calculate features for datapoints.
    `**kwargs` will get passed directly to `Featurizer.featurize`
    Parameters
    ----------
    datapoints: Iterable[Any]
      Any blob of data you like. Subclasss should instantiate this.
    """
    return self.featurize(datapoints, **kwargs)

  def _featurize(self, datapoint: Any, **kwargs):
    """Calculate features for a single datapoint.
    Parameters
    ----------
    datapoint: Any
      Any blob of data you like. Subclass should instantiate this.
    """
    raise NotImplementedError('Featurizer is not defined.')

  def __repr__(self) -> str:
    """Convert self to repr representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
    CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
    >>> jt.descriptors.CGCNNFeaturizer()
    CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_info = ''
    for arg_name in args_names:
      value = self.__dict__[arg_name]
      # for str
      if isinstance(value, str):
        value = "'" + value + "'"
      # for list
      if isinstance(value, list):
        threshold = get_print_threshold()
        value = np.array2string(np.array(value), threshold=threshold)
      args_info += arg_name + '=' + str(value) + ', '
    return self.__class__.__name__ + '[' + args_info[:-2] + ']'

  def __str__(self) -> str:
    """Convert self to str representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
    CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
    >>> jt.descriptors.CGCNNFeaturizer()
    CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_num = len(args_names)
    args_default_values = [None for _ in range(args_num)]
    if args_spec.defaults is not None:
      defaults = list(args_spec.defaults)
      args_default_values[-len(defaults):] = defaults

    override_args_info = ''
    for arg_name, default in zip(args_names, args_default_values):
      if arg_name in self.__dict__:
        arg_value = self.__dict__[arg_name]
        # validation
        # skip list
        if isinstance(arg_value, list):
          continue
        if isinstance(arg_value, str):
          # skip path string
          if "\\/." in arg_value or "/" in arg_value or '.' in arg_value:
            continue
        # main logic
        if default != arg_value:
          override_args_info += '_' + arg_name + '_' + str(arg_value)
    return self.__class__.__name__ + override_args_info


class MolecularFeaturizer(Featurizer):
  """Abstract class for calculating a set of features for a
  molecule.
  The defining feature of a `MolecularFeaturizer` is that it
  uses SMILES strings and RDKit molecule objects to represent
  small molecules. All other featurizers which are subclasses of
  this class should plan to process input which comes as smiles
  strings or RDKit molecules.
  Child classes need to implement the _featurize method for
  calculating features for a single molecule.

  .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
     O'Reilly Media (2019):.
  .. [2] https://github.com/deepchem/deepchem

  Note
  ----
  The subclasses of this class require RDKit to be installed.
  """
  def featurize(self, datapoints, log_every_n=1000, **kwargs) -> np.ndarray:
    """Calculate features for molecules.
    Parameters
    ----------
    datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
      RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
      strings.
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    Returns
    -------
    features: np.ndarray
      A numpy array containing a featurized representation of `datapoints`.
    """
    try:
      from rdkit import Chem
      from rdkit.Chem import rdmolfiles
      from rdkit.Chem import rdmolops
      from rdkit.Chem.rdchem import Mol
    except ModuleNotFoundError:
      raise ImportError("This class requires RDKit to be installed.")

    if 'molecules' in kwargs:
      datapoints = kwargs.get("molecules")
      raise DeprecationWarning(
          'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
      )

    # Special case handling of single molecule
    if isinstance(datapoints, str) or isinstance(datapoints, Mol):
      datapoints = [datapoints]
    else:
      # Convert iterables to list
      datapoints = list(datapoints)

    features: list = []
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, mol in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        if isinstance(mol, str):
          # mol must be a RDKit Mol object, so parse a SMILES
          mol = Chem.MolFromSmiles(mol)
          # SMILES is unique, so set a canonical order of atoms
          new_order = rdmolfiles.CanonicalRankAtoms(mol)
          mol = rdmolops.RenumberAtoms(mol, new_order)

        features.append(self._featurize(mol, **kwargs))
      except Exception as e:
        if isinstance(mol, Chem.rdchem.Mol):
          mol = Chem.MolToSmiles(mol)
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %d, %s. Appending empty array", i,
              mol)
          logger.warning("Exception message: {}".format(e))
        features.append(np.array([]))

    return np.asarray(features)

  def featurize_dataframe(self, datapoints, log_every_n=1000, **kwargs) -> Any:
    """Calculate features for molecules.
    Parameters
    ----------
    datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
      RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
      strings.
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    Returns
    -------
    features: pd.Dataframe()
      A pandas Dataframe containing a featurized representation of `datapoints`.
    """
    try:
      from rdkit import Chem
      from rdkit.Chem import rdmolfiles
      from rdkit.Chem import rdmolops
      from rdkit.Chem.rdchem import Mol
    except ModuleNotFoundError:
      raise ImportError("This class requires RDKit to be installed.")

    if 'molecules' in kwargs:
      datapoints = kwargs.get("molecules")
      raise DeprecationWarning(
          'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
      )

    # Special case handling of single molecule
    if isinstance(datapoints, str) or isinstance(datapoints, Mol):
      datapoints = [datapoints]
    else:
      # Convert iterables to list
      datapoints = list(datapoints)

    features: list = []
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, mol in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):

      try:
        if isinstance(mol, str):
          # mol must be a RDKit Mol object, so parse a SMILES
          mol = Chem.MolFromSmiles(mol)
          # SMILES is unique, so set a canonical order of atoms
          new_order = rdmolfiles.CanonicalRankAtoms(mol)
          mol = rdmolops.RenumberAtoms(mol, new_order)
          ar = self._featurize_dataframe(mol, **kwargs)
        else:
          ar = self._featurize_dataframe(mol, **kwargs)
        features.append(ar)
      except Exception as e:
        if isinstance(mol, Chem.rdchem.Mol):
          mol = Chem.MolToSmiles(mol)
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %d, %s. Appending empty array", i,
              mol)
          logger.warning("Exception message: {}".format(e))
        features.append(np.array([]))
        # features.append(self._featurize_dataframe(mol, **kwargs))
    columns = self._get_column_names()
    # features = np.array(features)
    if columns == ['Sequence']:
      df = pd.DataFrame({'Sequence': features})
    elif columns == ['OneHotSequence']:
      df = pd.DataFrame({'OneHotSequence': features})
    elif columns == ['SmilesImage']:
      df = pd.DataFrame({'SmilesImage': features})
    elif columns == ['MACCSFingerprint']:
      df = pd.DataFrame({'MACCSFingerprint': features})
    elif columns == ['MolGanGraphs']:
      df = pd.DataFrame({'MolGanGraphs': features})
    elif columns == ['CoulombMatrixEig']:
      df = pd.DataFrame({'CoulombMatrixEig': features})
    else:
      df = pd.DataFrame(features, columns=columns)
    return df


class ParserFeaturizer(Featurizer):
  """Abstract class for calculating a set of features for one or
  a set of files. The defining feature of a `ParserFeaturizer`
  is that it uses files to generate descriptors. All other
  featurizers which are subclasses of this class
  should plan to process input which comes as a parser of a file or
  a folder with files. Child classes need to implement the
  _featurize method for calculating features for a single instance.
  """

  def featurize(self, log_every_n=1000, **kwargs) -> np.ndarray:
    """Calculate features for materials.
    Parameters
    ----------
    material: string
      Either the path of a pdb file or the path to the folder with many pdb files.
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    files: str
      Extention of the files - Type of representation of materials
    Returns
    -------
    features: np.ndarray
      A numpy array containing a featurized representation of `datapoints`.
    """

    features: list = []
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, mat in enumerate(tqdm(self.parser.parse(), desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        features.append(self._featurize(mat, **kwargs))

      except Exception as e:
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %d, %s. Appending empty array", i, mat)
          logger.warning("Exception message: {}".format(e))
        features.append(np.array([]))
    return np.asarray(features)

  def featurize_dataframe(self, log_every_n=1000, **kwargs) -> Any:
    """Calculate features for materials.
    Parameters
    ----------
    pdb: string
      Either the path of a pdb file or the path to the folder with many pdb files.
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    files: str
      Extention of the files - Type of representation of materials.

    Returns
    -------
    features: pd.Dataframe()
      A pandas Dataframe containing a featurized representation of `datapoints`.
    """

    features = pd.DataFrame()
    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False
    for i, mat in enumerate(tqdm(self.parser.parse(), desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        features = pd.concat([features, self._featurize_dataframe(mat, **kwargs)]).reset_index(drop=True).replace(np.nan, 0)

      except Exception as e:
        if config.verbose is True:
          logger.warning(
            "Failed to featurize datapoint %d, %s. Appending empty array", i)
          logger.warning("Exception message: {}".format(e))
        features = pd.concat([features, pd.DataFrame()]).reset_index(drop=True)

    return features


class MaterialFeaturizer(Featurizer):
  """
  Abstract class for calculating a set of features for an
  inorganic crystal composition.
  The defining feature of a `MaterialFeaturizer` is that it
  operates on 3D crystal materials.
  Inorganic crystal compositions are represented by Pymatgen composition
  objects. Featurizers for inorganic crystal compositions that are
  subclasses of this class should plan to process input which comes as
  Pymatgen composition objects.
  This class is abstract and cannot be invoked directly. You'll
  likely only interact with this class if you're a developer. Child
  classes need to implement the _featurize method for calculating
  features for a single crystal composition.
  Note
  ----
  Some subclasses of this class will require pymatgen and matminer to be
  installed.
  """

  def featurize(self,
                datapoints: Union[List, Iterable[str]],
                log_every_n: int = 1000,
                **kwargs) -> np.ndarray:
    """Calculate features for crystal compositions.
    Parameters
    ----------
    datapoints: Iterable[str]
      Iterable sequence of composition strings, e.g. "MoS2".
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    Returns
    -------
    features: np.ndarray
      A numpy array containing a featurized representation of
      `compositions`.
    """

    if not isinstance(datapoints, List):
      datapoints = [datapoints]

    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False

    features = []
    for i, item in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        features.append(self._featurize(item, **kwargs))
      except Exception as e:
        if config.verbose is True:
          logger.warning(
              "Failed to featurize datapoint %i. Appending empty array" % i, e)
        features.append(np.array([]))

    return np.asarray(features)

  def featurize_dataframe(self, datapoints: Union[List, Iterable[str]], log_every_n=1000, **kwargs) -> pd.DataFrame:
    """Calculate features for materials.
    Parameters
    ----------
    pdb: string
      Either the path of a pdb file or the path to the folder with many pdb files.
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.
    files: str
      Extention of the files - Type of representation of materials.

    Returns
    -------
    features: pd.Dataframe()
      A pandas Dataframe containing a featurized representation of `datapoints`.
    """

    if not isinstance(datapoints, List):
      datapoints = [datapoints]

    if config.verbose is False:
      disable_tq = True
    else:
      disable_tq = False

    datapoints = list(datapoints)
    features = pd.DataFrame()
    for i, item in enumerate(tqdm(datapoints, desc='Creating descriptors', disable=disable_tq)):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        features = pd.concat([features, self._featurize_dataframe(item, **kwargs)]).reset_index(drop=True)

      except Exception as e:
        if config.verbose is True:
          logger.warning(
            "Failed to featurize datapoint %d, %s. Appending empty array", i)
          logger.warning("Exception message: {}".format(e))
        features = pd.concat([features, pd.DataFrame()]).reset_index(drop=True)

    return features


def get_print_threshold() -> int:
  """Return the printing threshold for datasets.
  The print threshold is the number of elements from ids/tasks to
  print when printing representations of `Dataset` objects.
  Returns
  ----------
  threshold: int
    Number of elements that will be printed
  """
  return _print_threshold


import unittest
from jaqpotpy import Jaqpot
import pandas as pd
from jaqpotpy.models import MolecularModel


class TestModels(unittest.TestCase):

    def test_login(self):
        jaqpot = Jaqpot()
        jaqpot.login("pantelispanka", "kapan2")

    def test_fail_login(self):
        jaqpot = Jaqpot()
        jaqpot.login("pante", "asdf")

    def test_get_model(self):
        jaqpot = Jaqpot()
        jaqpot.login("pantelispanka", "kapan2")
        model = jaqpot.get_model_by_id("Yon4zUuavqY46Umu9AKp")

    def test_load_mol_mod(self):
        jaqpot = Jaqpot()
        jaqpot.set_api_key("..TBKGvyRK8VbwCKYEd06tbpV-Z3VhiTTBwsKOkcB9nB-ZFM31J6nEYALqi-PYO7rPX4bHeRiC83dbMGs9_OdtsiApz_ayi8QLA006CvRkuaag4SXNQqiFQ")
        model = MolecularModel().load_from_jaqpot(jaqpot, "id")

    def test_predict(self):
        data = {'LFG': [0.1, 0.2], 'EQP': [0.1, 0.2], 'NEQ': [0.1, 0.2], 'GAP': [0.1, 0.2]}
        df = pd.DataFrame.from_dict(data)
        jaqpot = Jaqpot()
        jaqpot.login("pantelispanka", "kapan2")
        df, predicts = jaqpot.predict(df, "Yon4zUuavqY46Umu9AKp")
        print(df)
        print(predicts)



import jaqpotpy.api.login as jaqlogin
import jaqpotpy.api.algorithms_api as alapi
import jaqpotpy.api.dataset_api as data_api
import jaqpotpy.api.models_api as models_api
import jaqpotpy.api.chempot_api as chempot_api
import jaqpotpy.api.task_api as task_api
import jaqpotpy.api.doa_api as doa_api
import jaqpotpy.helpers.jwt as jwtok
import jaqpotpy.helpers.helpers as help
import jaqpotpy.helpers.dataset_deserializer as ds
# from jaqpotpy.helpers.logging import ColoredFormatter
from jaqpotpy.helpers.logging import init_logger
import json
import jaqpotpy.api.feature_api as featapi
from jaqpotpy.helpers.serializer import JaqpotSerializer
# from tornado.ioloop import IOLoop
from jaqpotpy.entities.dataset import Dataset
from jaqpotpy.entities.meta import MetaInfo
from jaqpotpy.entities.featureinfo import FeatureInfo
import pandas as pd
import numpy as np
import http.client as http_client
import getpass
import time
import jaqpotpy.doa.doa as jha
from sys import getsizeof
from tqdm import tqdm


ENCODING = 'utf-8'

# import matplotlib
# matplotlib.use('TkAgg')
# import matplotlib.pyplot as plt


class Jaqpot:
    """
    Deploys sklearn models on Jaqpot.

    Extended description of function.

    Parameters
    ----------
    base_url : The url on which Jaqpot services are deployed

    """

    def __init__(self, base_url=None, create_logs=False):
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        self.log = init_logger(__name__, testing_mode=False, output_log_file=create_logs)
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = 'https://api.jaqpot.org/jaqpot/services/'
        self.api_key = None
        self.user_id = None
        self.http_client = http_client

    def login(self, username, password):
        """
        Logins on Jaqpot.

        Parameters
        ----------
        username : username
        password : password

        """
        try:
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Could not login to jaqpot")
            # print("Error: " + str(e))

    def set_api_key(self, api_key):
        """
        Set's api key for authentication on Jaqpot.

        Parameters
        ----------
        api_key : api_key can be retireved from the application after logged in

        """
        self.api_key = api_key
        self.log.info("api key is set")

    def request_key(self, username, password):
        """
        Logins on Jaqpot.

        Parameters
        ----------
        username : username
        password : password

        """
        try:
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            self.log.info("api key is set")
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Error: " + str(e))

    def request_key_safe(self):
        """
        Logins on Jaqpot by hiding the users password.

        """
        try:
            username = input("Username: ")
            password = getpass.getpass("Password: ")
            au_req = jaqlogin.authenticate_sync(self.base_url, username, password)
            self.api_key = au_req['authToken']
            self.log.info("api key is set")
            # self.user_id = jwtok.decode_jwt(self.api_key).get('sub')
        except Exception as e:
            self.log.error("Error: " + str(e))

    def my_info(self):
        """
        Prints user's info

        """
        self.log.info(jwtok.decode_jwt(self.api_key))
        return jwtok.decode_jwt(self.api_key)

    def get_algorithms(self, start=None, max=None):
        self.check_key()
        try:
            algos = alapi.get_allgorithms_sync(self.base_url, self.api_key, start, max)
            return algos
        except Exception as e:
            self.log.error("Error:" + str(e))

    def get_algorithms_classes(self, start=None, max=None):
        self.check_key()
        try:
            algos = alapi.get_allgorithms_classes(self.base_url, self.api_key, start, max)
            return algos
        except Exception as e:
            self.log.error("Error:" + str(e))

    def upload_dataset(self, df=None, id=None, title=None, description=None):
        if title is None:
            raise Exception("Please submit title of the dataset")
        if description is None:
            raise Exception("Please submit description of the dataset")
        df_titles = list(df)
        for t in df_titles:
            type_to_c = df[t].dtypes
            if type_to_c == 'int64':
                # print(type_to_c)
                df[t] = df[t].astype(float)
        # print(df.dtypes)
        df = df.replace(np.nan, '', regex=True)
        if type(df).__name__ != 'DataFrame':
            raise Exception("Cannot form a Jaqpot Dataset. Please provide a Dataframe")
        if id is not None:
            df.set_index(id, inplace=True)
            feat_titles = list(df)
            feats = []
            for f_t in feat_titles:
                fe = help.create_feature(f_t, self.user_id)
                jsonmi = json.dumps(fe, cls=JaqpotSerializer)
                # jsonmi = json.dumps(fe.__dict__)
                feats.append(jsonmi)
        else:
            feat_titles = list(df)
            feats = []
            for f_t in feat_titles:
                fe = help.create_feature(f_t, self.user_id)
                # jsonmi = json.dumps(fe.__dict__)
                jsonmi = json.dumps(fe, cls=JaqpotSerializer)
                feats.append(jsonmi)
        feat_map = {}
        featutes = []
        keyi = 0
        for feat in feats:
            feat_info = FeatureInfo()
            f = featapi.create_feature_sync(self.base_url, self.api_key, feat)
            feat_uri = self.base_url + "feature/" + f["_id"]
            feat_map[f["meta"]["titles"][0]] = keyi
            feat_info.uri = feat_uri
            feat_info.name = f["meta"]["titles"][0]
            feat_info.key = keyi
            featutes.append(feat_info.__dict__)
            keyi += 1
        dataset = Dataset()
        meta = MetaInfo()
        meta.creators = [self.user_id]
        if title is not None:
            meta.titles = [title]
        meta.descriptions = [description]
        dataset.meta = meta.__dict__
        dataset.totalRows = df.shape[0]
        dataset.totalColumns = df.shape[1]
        dataset.existence = "UPLOADED"
        data_entry = help.create_data_entry(df, feat_map, self.user_id)
        dataset.dataEntry = data_entry
        dataset.features = featutes
        jsondataset = json.dumps(dataset, cls=JaqpotSerializer)
        dataset_n = data_api.create_dataset_sync(self.base_url, self.api_key, jsondataset, self.log)
        self.log.info("Dataset created with id: " + dataset_n["_id"])
        return dataset_n["_id"]

    def chempot_predict(self, model, smiles, descriptors = "mordred", doa = False):
        """
        Makes a prediction from a Chempot model

        Parameters
        ----------
        model : string
            model is the id of the chempot model
        smiles : string
            The input SMILES.
        descriptors : string
            The descriptors of the model.
        doa : boolean
            Whether the Domain of Applicability will be taken into account
        
        Returns
        -------
        tuple
            The predicted dataset (first item) and the predicted feature (second item)
        """
        task = chempot_api.predict(self.base_url, self.api_key, model, smiles, descriptors, doa, self.log)
        
        percentange = 0
        taskid = task['_id']
        while percentange < 100:
            time.sleep(1)
            task = task_api.get_task(self.base_url, self.api_key, taskid)
            try:
                percentange = task['percentageCompleted']
            except KeyError:
                percentange = 0
            self.log.info("completed " + str(percentange))
        predictedDataset = task['resultUri']
        dar = predictedDataset.split("/")
        dataset = data_api.get_dataset(self.base_url, self.api_key, dar[len(dar)-1], self.log)
        df, predicts = ds.decode_predicted(dataset)
        return df, predicts


    def predict(self, df=None, modelId=None):
        df_titles = list(df)
        for t in df_titles:
            type_to_c = df[t].dtypes
            # if type_to_c == 'int64':
                # print(type_to_c)
                # df[t] = df[t].astype(float)
        # print(df.dtypes)
        df = df.replace(np.nan, '', regex=True)
        if type(df).__name__ != 'DataFrame':
            raise Exception("Cannot form a Jaqpot Dataset. Please provide a Dataframe")
        model = models_api.get_model(self.base_url, self.api_key, modelId, self.log)
        # feats = []
        # for featUri in model['independentFeatures']:
        #     featar = featUri.split("/")
        #     feat = featapi.get_feature(self.base_url, self.api_key, featar[len(featar)-1])
        #     feats.append(feat)
        # for featUri in model['additionalInfo']['independentFeatures']:
        #     print(featUri)
            # featar = featUri.split("/")
            # feat = featapi.get_feature(self.base_url, self.api_key, featar[len(featar)-1])
            # feats.append(feat)
        feat_map = {}
        featutes = []
        keyi = 0

        for key,value in model['additionalInfo']['independentFeatures'].items():
            feat_info = FeatureInfo()
            # f = featapi.create_feature_sync(self.base_url, self.api_key, feat)
            feat_uri = key
            feat_map[value] = keyi
            feat_info.uri = feat_uri
            feat_info.name = value
            feat_info.key = keyi
            featutes.append(feat_info.__dict__)
            keyi += 1

        dataset = Dataset()
        meta = MetaInfo()
        meta.creators = [self.user_id]
        dataset.meta = meta.__dict__
        dataset.totalRows = df.shape[0]
        dataset.totalColumns = df.shape[1]
        # dataset.existence = "UPLOADED"
        data_entry = help.create_data_entry(df, feat_map, self.user_id)
        dataset.dataEntry = data_entry
        dataset.features = featutes
        jsondataset = json.dumps(dataset, cls=JaqpotSerializer)
        dataset_n = data_api.create_dataset_sync(self.base_url, self.api_key, jsondataset, self.log)
        datasetId = dataset_n["_id"]
        datasetUri = self.base_url + "dataset/" + datasetId
        task = models_api.predict(self.base_url, self.api_key, dataseturi=datasetUri, modelid=modelId, logger=self.log)
        percentange = 0
        taskid = task['_id']
        while percentange < 100:
            time.sleep(1)
            task = task_api.get_task(self.base_url, self.api_key, taskid)
            try:
                percentange = task['percentageCompleted']
            except KeyError:
                percentange = 0
            self.log.info("completed " + str(percentange))
        predictedDataset = task['resultUri']
        dar = predictedDataset.split("/")
        dataset = data_api.get_dataset(self.base_url, self.api_key, dar[len(dar)-1], self.log)
        df, predicts = ds.decode_predicted(dataset)
        return df, predicts

    def deploy_sklearn(self, model, X, y, title, description, model_meta=False, doa=None):
        """
        Deploys sklearn model or pipeline to Jaqpot.

        Extended description of function.

        Parameters
        ----------
        model : sklearn trained model
            model is a trained model that occurs from the sklearn.linear_model family of algorithms
        X : pandas dataframe
            The dataframe that is used to train the model (X variables).
        y : pandas dataframe
            The dataframe that is used to train the model (y variables).
        title: String
            The title of the model
        description: String
            The description of the model
        doa: pandas dataframe
            The dataset used to create the domain of applicability of the model

        Returns
        -------
        string
            The id of the model that uploaded
        """

        algorithm = type(model).__name__
        sk_version = model.__getstate__()['_sklearn_version']

        splited = sk_version.split(".")
        if int(splited[1]) < 21 and int(splited[0]) == 0:
            runtime = "scikit-learn-legacy"
        if int(splited[1]) == 22:
            runtime = "scikit-learn-0-22"
        if int(splited[1]) == 23:
            runtime = "scikit-learn-0-23"
        if int(splited[1]) == 24:
            runtime = "scikit-learn-0-24"
        if int(splited[1]) > 24:
            runtime = "scikit-learn-0-24"
        if int(splited[0]) == 1 and int(splited[1]) == 0:
            runtime = "scikit-learn-1-0"
        if int(splited[0]) == 1 and int(splited[1]) == 1:
            runtime = "scikit-learn-1-1"
        if int(splited[0]) == 1 and int(splited[1]) == 2:
            runtime = "scikit-learn-1-1"

        if isinstance(X, pd.DataFrame) is False and isinstance(X, pd.Series) is False:
            raise Exception('Function deploy_sklearn supports pandas dataframe or series. X is not one')
        if isinstance(y, pd.DataFrame) is False and isinstance(y, pd.Series) is False:
            raise Exception('Function deploy_sklearn supports pandas dataframe or series. Y is not one')

        additionalInfo = {}
        if model_meta is True:
            additionalInfo['meta'] = model.__getstate__()
        else:
            additionalInfo['meta'] = {}
            additionalInfo['meta']['_sklearn_version'] = sk_version
        additionalInfo['meta']['algorithm'] = algorithm
        if isinstance(X, pd.DataFrame):
            additionalInfo['inputSeries'] = list(X)
        if isinstance(X, pd.Series):
            additionalInfo['inputSeries'] = X.name
        pretrained = help.create_pretrain_req(model, X, y, title, description,
                                              algorithm, "Scikit learn model or pipeline", runtime,
                                              additionalInfo)
        size = getsizeof(pretrained.rawModel[0])
        if size < 2000000:
            j = json.dumps(pretrained, cls=JaqpotSerializer)
            if doa is None:
                response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
                if response.status_code < 300:
                    resp = response.json()
                    self.log.info("Model with id: " + resp['modelId'] + " created. Please visit the application to proceed")
                    return resp['modelId']
                else:
                    resp = response.json()
                    self.log.error("Some error occured: " + resp['message'])
                    return
            else:
                response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
                if response.status_code < 300:
                    resp = response.json()
                    self.log.info("Model with id: " + resp['modelId'] + " created. Storing Domain of applicability")
                    modid = resp['modelId']
                else:
                    resp = response.json()
                    self.log.error("Some error occured: " + resp['message'])
                    return
                # loop = asyncio.get_event_loop()
                # a = loop.create_task(jha.calculate_a(X))
                # b = loop.create_task(jha.calculate_doa_matrix(X))
                a = jha.calculate_a(X)
                b = jha.calculate_doa_matrix(X)
                # all_groups = asyncio.gather(a, b)
                # results = loop.run_until_complete(all_groups)
                doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp['modelId'])
                j = json.dumps(doa, cls=JaqpotSerializer)
                resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
                if resp == 201:
                    self.log.info("Stored Domain of applicability. Visit the application to proceed")
                    return modid
        else:
            rawModel = pretrained.rawModel[0]
            rawModel = str(rawModel)
            n = 1000000
            chunks = [rawModel[i:i + n] for i in range(0, len(rawModel), n)]
            del pretrained.rawModel
            j = json.dumps(pretrained, cls=JaqpotSerializer)
            if doa is None:
                response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
                if response.status_code < 300:
                    resp = response.json()
                    model_id = resp['modelId']
                    parts = []
                    i = 0
                    for c in chunks:
                        part = {}
                        part['modelId'] = model_id
                        part['partNumber'] = i
                        part['totalParts'] = len(chunks)
                        part['part'] = c
                        parts.append(part)
                        i += 1
                    for j in tqdm(range(len(chunks))):
                        part = parts[j]
                        json_r = json.dumps(part, cls=JaqpotSerializer)
                        resp = models_api.post_model_part(self.base_url, self.api_key, model_id, json_r, self.log)
                        # pass
                    # resp = response.json()
                    self.log.info("Model with id: " + model_id + " created. Please visit the application to proceed")
                    return model_id
                else:
                    resp = response.json()
                    self.log.error("Some error occured: " + resp['message'])
                    return
            else:
                response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
                resp = response.json()
                if response.status_code < 300:
                    resp = response.json()
                    model_id = resp['modelId']
                    self.log.info("Model with id: " + resp['modelId'] + " created. Storing Domain of applicability")
                    modid = resp['modelId']
                else:
                    resp = response.json()
                    self.log.error("Some error occured: " + resp['message'])
                    return
                a = jha.calculate_a(X)
                b = jha.calculate_doa_matrix(X)
                doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp['modelId'])
                j = json.dumps(doa, cls=JaqpotSerializer)
                resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
                if resp == 201:
                    self.log.info("Stored Domain of applicability. Visit the application to proceed")
                    return modid
            # print("Not supported")

    def deploy_jaqpotpy_molecular_model(self, model, description: str, title: str = None):
        if model.model_title != None:
            title = model.model_title
        pretrained = help.create_molecular_req(model, title=title, description=description, type="MolecularModel")
        size = getsizeof(pretrained.rawModel[0])
        if size < 2000000:
            j = json.dumps(pretrained, cls=JaqpotSerializer)
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Please visit the application to proceed")
                return resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
        else:
            rawModel = pretrained.rawModel[0]
            rawModel = str(rawModel)
            n = 2000000
            chunks = [rawModel[i:i + n] for i in range(0, len(rawModel), n)]
            del pretrained.rawModel
            j = json.dumps(pretrained, cls=JaqpotSerializer)
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                model_id = resp['modelId']
                parts = []
                i = 0
                for c in chunks:
                    part = {}
                    part['modelId'] = model_id
                    part['partNumber'] = i
                    part['totalParts'] = len(chunks)
                    part['part'] = c
                    parts.append(part)
                    i += 1
                for j in tqdm(range(len(chunks))):
                    part = parts[j]
                    json_r = json.dumps(part, cls=JaqpotSerializer)
                    resp = models_api.post_model_part(self.base_url, self.api_key, model_id, json_r, self.log)
                    # pass
                # resp = response.json()
                self.log.info("Model with id: " + model_id + " created. Please visit the application to proceed")
                return model_id
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return

    def deploy_sklearn_unsupervised(self, model, X, title, description, model_meta=False, doa=None):
        """
        Deploys sklearn model or pipeline to Jaqpot.

        Extended description of function.

        Parameters
        ----------
        model : sklearn trained model
            model is a trained model that occurs from the sklearn.linear_model family of algorithms
        X : pandas dataframe
            The dataframe that is used to train the model (X variables).
        title: String
            The title of the model
        description: String
            The description of the model
        doa: pandas dataframe
            The dataset used to create the domain of applicability of the model

        Returns
        -------
        string
            The id of the model that uploaded
        """

        algorithm = type(model).__name__
        sk_version = model.__getstate__()['_sklearn_version']

        splited = sk_version.split(".")
        if int(splited[1]) < 21 and int(splited[0]) == 0:
            runtime = "scikit-learn-legacy"
        if int(splited[1]) == 22:
            runtime = "scikit-learn-0-22"
        if int(splited[1]) == 23:
            runtime = "scikit-learn-0-23"
        if int(splited[1]) == 24:
            runtime = "scikit-learn-0-24"
        if int(splited[1]) > 24:
            runtime = "scikit-learn-0-24"
        if int(splited[0]) == 1 and int(splited[1]) == 0:
            runtime = "scikit-learn-1-0"
        if int(splited[0]) == 1 and int(splited[1]) == 1:
            runtime = "scikit-learn-1-1"

        if isinstance(X, pd.DataFrame) is False:
            raise Exception('Function deploy_glm supports pandas dataframe or series. X is not one')
        # if isinstance(y, pd.DataFrame) is False and isinstance(y, pd.Series) is False:
        #     raise Exception('Function deploy_glm supports pandas dataframe or series. Y is not one')

        y = pd.Series(data=None, index=['y'])

        additionalInfo = {}
        if model_meta is True:
            additionalInfo['meta'] = model.__getstate__()
        else:
            additionalInfo['meta'] = {}
            additionalInfo['meta']['_sklearn_version'] = sk_version
        additionalInfo['meta']['algorithm'] = algorithm
        if isinstance(X, pd.DataFrame):
            additionalInfo['inputSeries'] = list(X)
        if isinstance(X, pd.Series):
            additionalInfo['inputSeries'] = X.name
        pretrained = help.create_pretrain_req(model, X, y, title, description,
                                              algorithm, "Scikit learn model or pipeline", runtime,
                                              additionalInfo)
        j = json.dumps(pretrained, cls=JaqpotSerializer)
        if doa is None:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Please visit the application to proceed")
                return resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
        else:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Storing Domain of applicability")
                modid = resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
            # loop = asyncio.get_event_loop()
            # a = loop.create_task(jha.calculate_a(X))
            # b = loop.create_task(jha.calculate_doa_matrix(X))
            # all_groups = asyncio.gather(a, b)
            # results = loop.run_until_complete(all_groups)
            a = jha.calculate_a(X)
            b = jha.calculate_doa_matrix(X)
            doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp['modelId'])
            j = json.dumps(doa, cls=JaqpotSerializer)
            resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
            if resp == 201:
                self.log.info("Stored Domain of applicability. Visit the application to proceed")
                return modid

    def api_key(self):
        return self.api_key

    def check_key(self):
        if self.api_key is None:
            raise Exception("Api key not present!")

    def deploy_XGBoost(self, model, X, y, title, description, algorithm, doa=None):
        """
        Deploys XGBoost model to Jaqpot.

        Extended description of function.

        Parameters
        ----------
        model : XGBoost trained model
            model is a trained model that occurs from the XGBoost family of algorithms
        X : pandas dataframe
            The dataframe that is used to train the model (X variables).
        y : pandas dataframe
            The dataframe that is used to train the model (y variables).
        title: String
            The title of the model
        description: String
            The description of the model
        algorithm: String
            The algorithm that the model implements
        doa: pandas dataframe
            The dataset used to create the domain of applicability of the model

        Returns
        -------
        string
            The id of the model that uploaded

        """
        if isinstance(X, pd.DataFrame) is False:
            raise Exception('Function deploy_glm supports pandas dataframe or series. X is not one')
        if isinstance(y, pd.DataFrame) is False and isinstance(y, pd.Series) is False:
            raise Exception('Function deploy_glm supports pandas dataframe or series. Y is not one')
        # coef_flatten = model.coef_.flatten()
        # intercept_flatten = model.intercept_.flatten()
        # coef_df = pd.DataFrame(list(zip(list(X), coef_flatten)), columns=['Features', 'coeff'])
        # intercept = pd.DataFrame([('(intercept)', intercept_flatten[0])], columns=['Features', 'coeff'])
        # coefs_all = intercept.append(coef_df, ignore_index=True)
        # coefs = {}
        additionalInfo = {}
        # for key in coefs_all.values:
        #     coefs[key[0]] = key[1]
        # additionalInfo['coefficients'] = coefs
        additionalInfo['inputSeries'] = list(X)
        pretrained = help.create_pretrain_req(model, X, y, title, description,
                                              algorithm, "XGBoost model", "XGBoost-model",
                                              additionalInfo)
        # j = json.dumps(pretrained)
        j = json.dumps(pretrained, cls=JaqpotSerializer)
        if doa is None:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Please visit the application to proceed")
                return resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
        else:
            response = models_api.post_pretrained_model(self.base_url, self.api_key, j, self.log)
            if response.status_code < 300:
                resp = response.json()
                self.log.info("Model with id: " + resp['modelId'] + " created. Storing Domain of applicability")
                modid = resp['modelId']
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp['message'])
                return
            # loop = asyncio.get_event_loop()
            # a = loop.create_task(jha.calculate_a(X))
            # b = loop.create_task(jha.calculate_doa_matrix(X))

            a = jha.calculate_a(X)
            b = jha.calculate_doa_matrix(X)
            # all_groups = asyncio.gather(a, b)
            # results = loop.run_until_complete(all_groups)
            doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp['modelId'])
            j = json.dumps(doa, cls=JaqpotSerializer)
            resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
            if resp == 201:
                self.log.info("Stored Domain of applicability. Visit the application to proceed")
                return modid

    def get_model_by_id(self, model):
        """
        Retrieves user's model by ID.

        Parameters
        ----------
        model : str
            The model's ID.
        
        
        Returns
        -------
        Object
            The particular model.

        """
        return models_api.get_model(self.base_url, self.api_key, model, self.log)

    def get_raw_model_by_id(self, model):
        """
        Retrieves raw model by ID.

        Parameters
        ----------
        model : str
            The model's ID.

        Returns
        Returns
        -------
        Object
            The particular model and the raw model.
        """

        # raw_model = models_api.get_raw_model(self.base_url, self.api_key, model, self.log)

        validating = jaqlogin.validate_api_key(self.base_url, self.api_key)
        if validating is not True:
            self.log.error(validating)
        else:
            return models_api.get_raw_model(self.base_url, self.api_key, model, self.log)

    def get_feature_by_id(self, feature):
        """
        Retrieves a Jaqpot feature.

        Parameters
        ----------
        feature : str
            The feature's ID.
        
        
        Returns
        -------
        Object
            The particular feature.

        """
        return featapi.get_feature(self.base_url, self.api_key, feature)

    def get_my_models(self, minimum, maximum):
        """
        Retrieves user's models.

        Parameters
        ----------
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the user.

        """
        return models_api.get_my_models(self.base_url, self.api_key, minimum, maximum, self.log)

    
    def get_orgs_models(self, organization, minimum, maximum):
        """
        Retrieves organization's models.

        Parameters
        ----------
        organization: str
            The organization's ID.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_orgs_models(self.base_url, self.api_key, organization, minimum, maximum, self.log)


    def get_models_by_tag(self, tag, minimum, maximum):
        """
        Retrieves models with a particular tag.

        Parameters
        ----------
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(self.base_url, self.api_key, tag, minimum, maximum, self.log)

    def get_models_by_tag_and_org(self, organization, tag, minimum, maximum):
        """
        Retrieves models of an organization with a particular tag.

        Parameters
        ----------
        organization: str:
            The organization's ID.
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.
        
        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(self.base_url, self.api_key, organization, tag, minimum, maximum, self.log)


    def get_dataset(self, dataset):
        """
        Retrieves a dataset.

        Parameters
        ----------
        dataset: str
            The dataset's ID.
        
        Returns
        -------
        pd.DataFrame()
            DataFrame of the dataset.

        Note
        ----
        Mazimum rows retrieved: 1000.
        """

        dataRetrieved = data_api.get_dataset(self.base_url, self.api_key, dataset, self.log)

        feats = ["" for i in range(len(dataRetrieved['features']))]
        
        for item in dataRetrieved['features']:
            feats[int(item['key'])] = item['name']

        indices = [item['entryId']['name'] for item in dataRetrieved['dataEntry']]

        data = [[item['values'][str(feats.index(col))] for col in feats] for item in dataRetrieved['dataEntry']]

        return pd.DataFrame(data,index=indices, columns=feats) 

    def get_doa(self,model):
        """
        Retrieves model's domain of applicability.

        Parameters
        ----------
        model: str
            The model's ID.
        
        Returns
        -------
        Object
            The model's domain of applicability.

        """
        return doa_api.get_models_doa(self.base_url, self.api_key, model, self.log)



from jaqpotpy.models.base_classes import Model
from jaqpotpy.doa.doa import DOA
from typing import Any
from jaqpotpy.datasets import TorchGraphDataset
from jaqpotpy.models import Evaluator, Preprocesses
try:
    from torch_geometric.loader import DataLoader
    import torch_geometric
except ModuleNotFoundError:
    DataLoader = None
    pass
import torch
from jaqpotpy.models import MolecularModel
import numpy as np
from jaqpotpy.cfg import config
import os
import jaqpotpy


# class MolecularTorchGeometric(Model):
#
#     def __init__(self, dataset: TorchGraphDataset, model_nn: torch.nn.Module
#                  , doa: DOA = None
#                  , eval: Evaluator = None, preprocess: Preprocesses = None
#                  , dataLoaderParams: Any = None, epochs: int = None
#                  , criterion: torch.nn.Module = None, optimizer: Any = None
#                  , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1, model_dir: str = "./", device: str = 'cpu'):
#         # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
#         self.dataset: TorchGraphDataset = dataset
#         self.model_nn = model_nn
#         self.doa = doa
#         self.doa_m = None
#         self.external = None
#         self.evaluator: Evaluator = eval
#         self.preprocess: Preprocesses = preprocess
#         self.train_batch = train_batch
#         self.test_batch = test_batch
#         self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
#         self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
#         self.epochs = epochs
#         self.criterion = criterion
#         self.trained_model = None
#         self.model_fitted: MolecularModel = None
#         self.optimizer_local = optimizer
#         self.train_loader = None
#         self.test_loader = None
#         self.log_steps = log_steps
#         self.best_model = model_nn
#         self.path = None
#         self.model_dir = model_dir
#         self.device = torch.device(device)
#         # torch.multiprocessing.freeze_support()
#
#     def __call__(self, smiles):
#         self
#
#     def fit(self):
#         steps = self.epochs * 0.1
#         if self.doa:
#             if self.doa.__name__ == 'SmilesLeverage':
#                 self.doa_m = self.doa.fit(self.dataset.smiles)
#             else:
#                 print("Only SmilesLeverage is suported for graph models")
#         if len(self.dataset.df) > 0:
#             pass
#         else:
#             self.dataset.create()
#         if self.evaluator is not None:
#             if len(self.evaluator.dataset.df) > 0:
#                 pass
#             else:
#                 self.evaluator.dataset.create()
#         self.model_fitted = MolecularModel()
#         self.model_nn.to(self.device)
#         # self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
#         # self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
#         self.train_loader = DataLoader(dataset=self.dataset, **self.trainDataLoaderParams)
#         self.test_loader = DataLoader(dataset=self.evaluator.dataset, **self.testDataLoaderParams)
#         temp_loss = None
#         for epoch in range(1, self.epochs):
#             self.train()
#             train_loss = self.test(self.train_loader)
#             test_loss = self.test(self.test_loader)
#             if epoch % self.log_steps == 0:
#                 if self.dataset.task == "classification":
#                     los = test_loss[2]
#                     if temp_loss is None or temp_loss < los:
#                         temp_loss = los
#                         temp_path = self.path
#                         self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(los) + ".pt"
#                         torch.save({
#                             'epoch': epoch,
#                             'model_state_dict': self.model_nn.state_dict(),
#                             'optimizer_state_dict': self.optimizer_local.state_dict(),
#                             'loss': los,
#                         }, self.path)
#                         try:
#                             os.remove(temp_path)
#                         except TypeError as e:
#                             continue
#                     print(f'Epoch: {epoch:03d}, Train Accuracy: {train_loss[2]}, Test Accuracy: {test_loss[2]}')
#                 else:
#                     los = test_loss[1]
#                     if temp_loss is None or temp_loss > los:
#                         temp_loss = los
#                         temp_path = self.path
#                         # print(temp_loss.item())
#                         # print(str(temp_loss))
#                         self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(temp_loss.item()) + ".pt"
#                         torch.save({
#                             'epoch': epoch,
#                             'model_state_dict': self.model_nn.state_dict(),
#                             'optimizer_state_dict': self.optimizer_local.state_dict(),
#                             'loss': los,
#                         }, self.path)
#                         try:
#                             os.remove(temp_path)
#                         except TypeError as e:
#                             continue
#                     print(f'Epoch: {epoch:03d}, Train Loss: {train_loss[1]}, Test Loss: {test_loss[1]}')
#         return self
#
#     def train(self):
#         self.model_nn.train()
#         for data in self.train_loader:
#             out = self.model_nn(data)
#             loss = self.criterion(out, data.y)
#             loss.backward()
#             self.optimizer_local.step()
#             self.optimizer_local.zero_grad()
#
#     def test(self, dataloader):
#         self.model_nn.eval()
#         if self.dataset.task == 'classification':
#             correct = 0
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#             return out, pred.numpy(), correct / len(dataloader.dataset)
#         else:
#             self.model_nn.eval()
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 loss = self.criterion(out, data.y)
#             return out, loss
#
#     def eval(self):
#         checkpoint = torch.load(self.path)
#         self.best_model.load_state_dict(checkpoint['model_state_dict'])
#         self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
#         self.best_model.eval()
#         if self.evaluator.functions:
#             eval_keys = self.evaluator.functions.keys()
#             correct = 0
#             truth = np.array([])
#             preds = np.array([])
#             for data in self.test_loader:
#                 out = self.best_model(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#                 truth = np.append(truth, data.y.numpy())
#                 preds = np.append(preds, pred.numpy())
#             # return out, pred.numpy(), correct / len(dataloader.dataset)
#             for eval_key in eval_keys:
#                 eval_function = self.evaluator.functions.get(eval_key)
#                 print(eval_key + ": " + str(eval_function(truth, preds)))
#         else:
#             print("No eval functions passed")
#
#     def create_molecular_model(self):
#         checkpoint = torch.load(self.path)
#         self.best_model.load_state_dict(checkpoint['model_state_dict'])
#         self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
#         model = MolecularModel()
#         model.descriptors = self.dataset.featurizer
#         model.doa = self.doa
#         model.model = self.best_model
#         model.optimizer = self.optimizer_local
#         model.X = self.dataset.X
#         model.Y = self.dataset.y
#         model.library = ['torch_geometric', 'torch']
#         model.version = [torch_geometric.__version__, torch.__version__]
#         model.jaqpotpy_version = jaqpotpy.__version__
#         model.jaqpotpy_docker = config.jaqpotpy_docker
#         model.modeling_task = self.dataset.task
#         return model


class MolecularTorchGeometric(Model):

    def __init__(self, dataset: TorchGraphDataset, model_nn: torch.nn.Module
                 , doa: DOA = None
                 , eval: Evaluator = None, preprocess: Preprocesses = None
                 , dataLoaderParams: Any = None, epochs: int = None
                 , criterion: torch.nn.Module = None, optimizer: Any = None
                 , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1, model_dir: str = "./", device: str = 'cpu', test_metric=(None, 'minimize')):
        # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
        self.dataset: TorchGraphDataset = dataset
        self.model_nn = model_nn
        self.doa = doa
        self.doa_m = None
        self.external = None
        self.evaluator: Evaluator = eval
        self.preprocess: Preprocesses = preprocess
        self.train_batch = train_batch
        self.test_batch = test_batch
        self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
        self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
        self.epochs = epochs
        self.criterion = criterion
        self.trained_model = None
        self.model_fitted: MolecularModel = None
        self.optimizer_local = optimizer
        self.train_loader = None
        self.test_loader = None
        self.log_steps = log_steps
        self.best_model = model_nn
        self.path = None
        self.model_dir = model_dir
        self.device = torch.device(device)
        self.__default__ = False
        if test_metric[0] is None and dataset.task == 'classification':
            self.__default__ = True
            test_metric = ('Accuracy', 'maximize')
        if test_metric[0] is None and dataset.task == 'regression':
            self.__default__ = True
            test_metric = ('Loss', 'minimize')
        self.test_metric = test_metric

        # torch.multiprocessing.freeze_support()

    def __call__(self, smiles):
        self

    def fit(self):
        steps = self.epochs * 0.1
        if self.doa:
            if self.doa.__name__ == 'SmilesLeverage':
                self.doa_m = self.doa.fit(self.dataset.smiles)
            else:
                print("Only SmilesLeverage is suported for graph models")
        if len(self.dataset.df) > 0:
            pass
        else:
            self.dataset.create()
        if self.evaluator is not None:
            if len(self.evaluator.dataset.df) > 0:
                pass
            else:
                self.evaluator.dataset.create()
        self.model_fitted = MolecularModel()
        self.model_nn.to(self.device)
        # self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
        # self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
        self.train_loader = DataLoader(dataset=self.dataset, **self.trainDataLoaderParams)
        self.test_loader = DataLoader(dataset=self.evaluator.dataset, **self.testDataLoaderParams)
        temp_loss = None
        for epoch in range(1, self.epochs):
            self.train()
            train_loss = self.test(self.train_loader)
            test_loss = self.test(self.test_loader)
            if epoch % self.log_steps == 0:
                if self.dataset.task == "classification":
                    los = test_loss[2]
                    if self.__save_choice__(temp_loss, train_loss[2], los):
                        temp_loss = [train_loss[2], los]
                        temp_path = self.path
                        self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(los) + ".pt"
                        torch.save({
                            'epoch': epoch,
                            'model_state_dict': self.model_nn.state_dict(),
                            'optimizer_state_dict': self.optimizer_local.state_dict(),
                            'loss': los,
                        }, self.path)
                        try:
                            os.remove(temp_path)
                        except TypeError as e:
                            continue
                    print(f'Epoch: {epoch:03d}, Train Score: {train_loss[2]}, Test Score: {test_loss[2]}')
                else:
                    los = test_loss[1]
                    if self.__save_choice__(temp_loss, train_loss[1], los):
                        temp_loss = [train_loss[1], los]
                        temp_path = self.path
                        # print(temp_loss.item())
                        # print(str(temp_loss))
                        self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(los.item()) + ".pt"
                        torch.save({
                            'epoch': epoch,
                            'model_state_dict': self.model_nn.state_dict(),
                            'optimizer_state_dict': self.optimizer_local.state_dict(),
                            'loss': los,
                        }, self.path)
                        try:
                            os.remove(temp_path)
                        except TypeError as e:
                            continue
                    print(f'Epoch: {epoch:03d}, Train Score: {train_loss[1]}, Test Score: {test_loss[1]}')
        return self


    def __save_choice__(self, temp_loss, train_loss, test_loss):
        if self.test_metric[1] == 'minimize':
            if temp_loss is None:
                return True
            elif temp_loss[1] > test_loss:
                return True
            elif temp_loss[1] == test_loss:
                if temp_loss[0] >= train_loss:
                    return True
                else:
                    return False
            else:
                return False
        elif self.test_metric[1] == 'maximize':
            if temp_loss is None:
                return True
            elif temp_loss[1] < test_loss:
                return True
            elif temp_loss[1] == test_loss:
                if temp_loss[0] <= train_loss:
                    return True
                else:
                    return False
            else:
                return False

    def train(self):
        self.model_nn.train()
        for data in self.train_loader:
            x = data.x.to(self.device)
            edge_index = data.edge_index.to(self.device)
            try:
                edge_attributes = data.edge_attr.to(self.device)
            except AttributeError:
                edge_attributes = None
            batch = data.batch.to(self.device)
            y = data.y.to(self.device)
            if edge_attributes is not None:

                out = self.model_nn(x, edge_index, edge_attributes, batch)
            else:
                out = self.model_nn(x, edge_index, batch)
            loss = self.criterion(out, y)
            loss.backward()
            self.optimizer_local.step()
            self.optimizer_local.zero_grad()

    def test(self, dataloader):
        if self.__default__:
            return self.__default_test__(dataloader)
        else:
            self.model_nn.eval()
            test_function = self.test_metric[0]
            truth = np.array([])
            preds = np.array([])
            for data in dataloader:
                x = data.x.to(self.device)
                edge_index = data.edge_index.to(self.device)
                try:
                    edge_attributes = data.edge_attr.to(self.device)
                except AttributeError:
                    edge_attributes = None
                batch = data.batch.to(self.device)

                y = data.y.to(self.device)

                if edge_attributes is not None:
                    out = self.model_nn(x, edge_index, edge_attributes, batch)
                else:
                    out = self.model_nn(x, edge_index, batch)

                truth = np.append(truth, y.cpu().numpy())

                if self.dataset.task == "regression":
                    preds = np.append(preds, out.cpu().detach().numpy())
                    return out, test_function(truth, preds)
                else:
                    preds = np.append(preds, out.argmax(dim=1).cpu().detach().numpy())
                    return out, out.argmax(dim=1).cpu().numpy(), test_function(truth, preds)


    def __default_test__(self, dataloader):
        self.model_nn.eval()
        if self.dataset.task == 'classification':
            correct = 0
            for data in dataloader:
                x = data.x.to(self.device)
                edge_index = data.edge_index.to(self.device)
                try:
                    edge_attributes = data.edge_attr.to(self.device)
                except AttributeError:
                    edge_attributes = None
                batch = data.batch.to(self.device)

                y = data.y.to(self.device)

                if edge_attributes is not None:
                    out = self.model_nn(x, edge_index, edge_attributes, batch)
                else:
                    out = self.model_nn(x, edge_index, batch)
                pred = out.argmax(dim=1)
                correct += int((pred == y).sum())
            return out, pred.cpu().numpy(), correct / len(dataloader.dataset)
        else:
            self.model_nn.eval()
            for data in dataloader:
                x = data.x.to(self.device)
                edge_index = data.edge_index.to(self.device)
                edge_attributes = None
                if data.edge_attr is not None:
                    edge_attributes = data.edge_attr.to(self.device)
                batch = data.batch.to(self.device)

                y = data.y.to(self.device)

                if edge_attributes is not None:
                    out = self.model_nn(x, edge_index, edge_attributes, batch)
                else:
                    out = self.model_nn(x, edge_index, batch)
                loss = self.criterion(out, y)
            return out, loss

    def eval(self):
        checkpoint = torch.load(self.path)
        self.best_model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
        self.best_model.eval()
        self.best_model.to(self.device)
        if self.evaluator.functions:
            eval_keys = self.evaluator.functions.keys()
            correct = 0
            truth = np.array([])
            preds = np.array([])
            for data in self.test_loader:
                x = data.x.to(self.device)
                edge_index = data.edge_index.to(self.device)
                edge_attributes = None
                if data.edge_attr is not None:
                    edge_attributes = data.edge_attr.to(self.device)
                batch = data.batch.to(self.device)

                y = data.y.to(self.device)

                if edge_attributes is not None:
                    out = self.best_model(x, edge_index, edge_attributes, batch)
                else:
                    out = self.best_model(x, edge_index, data.batch)
                if self.dataset.task == "regression":
                    pred = out
                else:
                    pred = out.argmax(dim=1)
                correct += int((pred == y).sum())
                truth = np.append(truth, y.cpu().numpy())
                preds = np.append(preds, pred.cpu().detach().numpy())
            # return out, pred.numpy(), correct / len(dataloader.dataset)
            for eval_key in eval_keys:
                eval_function = self.evaluator.functions.get(eval_key)
                print(eval_key + ": " + str(eval_function(truth, preds)))
        else:
            print("No eval functions passed")

    def create_molecular_model(self):
        checkpoint = torch.load(self.path)
        self.best_model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
        model = MolecularModel()
        model.descriptors = self.dataset.featurizer
        model.doa = self.doa
        # model.model = self.best_model
        self.best_model.to("cpu")
        model_s = torch.jit.script(self.best_model)
        model_save = torch.jit.save(model_s, "local_temp.pt")
        with open("./local_temp.pt", "rb") as f:
            model.model = f.read()
        f.close()
        os.remove("./local_temp.pt")
        # model.model = torch.jit.script(self.best_model)
        model.optimizer = self.optimizer_local
        model.X = self.dataset.X
        model.Y = self.dataset.y
        model.library = ['torch_geometric', 'torch']
        model.version = [torch_geometric.__version__, torch.__version__]
        model.jaqpotpy_version = jaqpotpy.__version__
        model.jaqpotpy_docker = config.jaqpotpy_docker
        model.modeling_task = self.dataset.task
        return model



# class MolecularTorch(Model):
#
#     def __init__(self, dataset: TorchGraphDataset, model_nn: torch.nn.Module
#                  , doa: DOA = None
#                  , eval: Evaluator = None, preprocess: Preprocesses = None
#                  , dataLoaderParams: Any = None, epochs: int = None
#                  , criterion: torch.nn.Module = None, optimizer: Any = None
#                  , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1):
#         # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
#         self.dataset: TorchGraphDataset = dataset
#         self.model_nn = model_nn
#         self.doa = doa
#         self.doa_m = None
#         self.external = None
#         self.evaluator: Evaluator = eval
#         self.preprocess: Preprocesses = preprocess
#         self.train_batch = train_batch
#         self.test_batch = test_batch
#         self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
#         self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
#         self.epochs = epochs
#         self.criterion = criterion
#         self.trained_model = None
#         self.model_fitted: MolecularModel = None
#         self.optimizer = optimizer
#         self.train_loader = None
#         self.test_loader = None
#         self.log_steps = log_steps
#         self.best_model = None
#         # torch.multiprocessing.freeze_support()
#
#     def __call__(self, smiles):
#         self
#
#     def fit(self):
#         if self.doa:
#             if self.doa.__name__ == 'SmilesLeverage':
#                 self.doa_m = self.doa.fit(self.dataset.smiles_strings)
#             else:
#                 print("Only SmilesLeverage is suported for graph models")
#         if self.dataset.df is not None:
#             pass
#         else:
#             self.dataset.create()
#         if self.evaluator is not None:
#             if self.evaluator.dataset.df is not None:
#                 pass
#             else:
#                 self.evaluator.dataset.create()
#         device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#         self.model_fitted = MolecularModel()
#         self.model_nn.to(device)
#         self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
#         self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
#         temp_loss = None
#         for epoch in range(1, self.epochs):
#             self.train()
#             train_loss = self.test(self.train_loader)
#             test_loss = self.test(self.test_loader)
#             if epoch % self.log_steps == 0:
#                 if self.dataset.task == "classification":
#                     los = test_loss[2]
#                     if temp_loss is None:
#                         temp_loss = test_loss[2]
#                     if temp_loss < los:
#                         self.best_model = self.model_nn
#                     print(f'Epoch: {epoch:03d}, Train Accuracy: {train_loss[2]}, Test Accuracy: {test_loss[2]}')
#                 else:
#                     los = test_loss[1]
#                     if temp_loss is None:
#                         temp_loss = test_loss[1]
#                     if temp_loss < los:
#                         self.best_model = self.model_nn
#                     print(f'Epoch: {epoch:03d}, Train Loss: {train_loss[1]}, Test Loss: {test_loss[1]}')
#         return self
#
#     def train(self):
#         self.model_nn.train()
#         for data in self.train_loader:
#             out = self.model_nn(data)
#             loss = self.criterion(out, data.y)
#             loss.backward()
#             self.optimizer.step()
#             self.optimizer.zero_grad()
#
#     def test(self, dataloader):
#         self.model_nn.eval()
#         if self.dataset.task == 'classification':
#             correct = 0
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#             return out, pred.numpy(), correct / len(dataloader.dataset)
#         else:
#             self.model_nn.eval()
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 loss = self.criterion(out, data.y)
#             return out, loss
#
#     def eval(self):
#         self.best_model.eval()
#         if self.evaluator.functions:
#             eval_keys = self.evaluator.functions.keys()
#             correct = 0
#             truth = np.array([])
#             preds = np.array([])
#             for data in self.test_loader:
#                 out = self.best_model(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#                 truth = np.append(truth, data.y.numpy())
#                 preds = np.append(preds, pred.numpy())
#             # return out, pred.numpy(), correct / len(dataloader.dataset)
#             for eval_key in eval_keys:
#                 eval_function = self.evaluator.functions.get(eval_key)
#                 print(eval_key + ":" + str(eval_function(truth, preds)))
#         else:
#             print("No eval functions passed")


# class MolecularTorch(Model):
#
#     def __init__(self, dataset: TorchGraphDataset, model_nn: torch.nn.Module
#                  , doa: DOA = None
#                  , eval: Evaluator = None, preprocess: Preprocesses = None
#                  , dataLoaderParams: Any = None, epochs: int = None
#                  , criterion: torch.nn.Module = None, optimizer: Any = None
#                  , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1):
#         # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
#         self.dataset: TorchGraphDataset = dataset
#         self.model_nn = model_nn
#         self.doa = doa
#         self.doa_m = None
#         self.external = None
#         self.evaluator: Evaluator = eval
#         self.preprocess: Preprocesses = preprocess
#         self.train_batch = train_batch
#         self.test_batch = test_batch
#         self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
#         self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
#         self.epochs = epochs
#         self.criterion = criterion
#         self.trained_model = None
#         self.model_fitted: MolecularModel = None
#         self.optimizer = optimizer
#         self.train_loader = None
#         self.test_loader = None
#         self.log_steps = log_steps
#         self.best_model = None
#         # torch.multiprocessing.freeze_support()
#
#     def __call__(self, smiles):
#         self
#
#     def fit(self):
#         if self.doa:
#             if self.doa.__name__ == 'SmilesLeverage':
#                 self.doa_m = self.doa.fit(self.dataset.smiles_strings)
#             else:
#                 print("Only SmilesLeverage is suported for graph models")
#         if self.dataset.df is not None:
#             pass
#         else:
#             self.dataset.create()
#         if self.evaluator is not None:
#             if self.evaluator.dataset.df is not None:
#                 pass
#             else:
#                 self.evaluator.dataset.create()
#         device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#         self.model_fitted = MolecularModel()
#         self.model_nn.to(device)
#         # self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
#         # self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
#         self.train_loader = DataLoader(dataset=self.dataset, **self.trainDataLoaderParams)
#         self.test_loader = DataLoader(dataset=self.evaluator.dataset, **self.testDataLoaderParams)
#         temp_loss = None
#         for epoch in range(1, self.epochs):
#             self.train()
#             train_loss = self.test(self.train_loader)
#             test_loss = self.test(self.test_loader)
#             if epoch % self.log_steps == 0:
#                 if self.dataset.task == "classification":
#                     los = test_loss[2]
#                     if temp_loss is None:
#                         temp_loss = test_loss[2]
#                     if temp_loss < los:
#                         self.best_model = self.model_nn
#                     print(f'Epoch: {epoch:03d}, Train Accuracy: {train_loss[2]}, Test Accuracy: {test_loss[2]}')
#                 else:
#                     los = test_loss[1]
#                     if temp_loss is None:
#                         temp_loss = test_loss[1]
#                     if temp_loss < los:
#                         self.best_model = self.model_nn
#                     print(f'Epoch: {epoch:03d}, Train Loss: {train_loss[1]}, Test Loss: {test_loss[1]}')
#         return self
#
#     def train(self):
#         self.model_nn.train()
#         for data in self.train_loader:
#             out = self.model_nn(data)
#             loss = self.criterion(out, data.y)
#             loss.backward()
#             self.optimizer.step()
#             self.optimizer.zero_grad()
#
#     def test(self, dataloader):
#         self.model_nn.eval()
#         if self.dataset.task == 'classification':
#             correct = 0
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#             return out, pred.numpy(), correct / len(dataloader.dataset)
#         else:
#             self.model_nn.eval()
#             for data in dataloader:
#                 out = self.model_nn(data)
#                 loss = self.criterion(out, data.y)
#             return out, loss
#
#     def eval(self):
#         self.best_model.eval()
#         if self.evaluator.functions:
#             eval_keys = self.evaluator.functions.keys()
#             correct = 0
#             truth = np.array([])
#             preds = np.array([])
#             for data in self.test_loader:
#                 out = self.best_model(data)
#                 pred = out.argmax(dim=1)
#                 correct += int((pred == data.y).sum())
#                 truth = np.append(truth, data.y.numpy())
#                 preds = np.append(preds, pred.numpy())
#             # return out, pred.numpy(), correct / len(dataloader.dataset)
#             for eval_key in eval_keys:
#                 eval_function = self.evaluator.functions.get(eval_key)
#                 print(eval_key + ":" + str(eval_function(truth, preds)))
#         else:
#             print("No eval functions passed")



from rdkit import DataStructs
from rdkit import Chem
from rdkit.Chem import QED
from rdkit.Chem import Crippen
from rdkit.Chem import AllChem
from rdkit.Chem import Draw
import pickle
import gzip
import numpy as np
import math

import pkg_resources
from typing import Iterable, Any

NP_file = pkg_resources.resource_filename('jaqpotpy.models.generative.data', 'NP_score.pkl.gz')
SA_file = pkg_resources.resource_filename('jaqpotpy.models.generative.data', 'SA_score.pkl.gz')

NP_model = pickle.load(gzip.open(NP_file))
SA_model = {i[j]: float(i[0]) for i in pickle.load(gzip.open(SA_file)) for j in range(1, len(i))}


def validate_molecule(mol):
    try:
        x = Chem.MolToSmiles(mol)
    except Exception as e:
        x = None
    return x


def novel_score(mols: Iterable[Any], smiles: Iterable[str]):
    valid_mols = []
    valid_smiles = []
    for mol in mols:
        smiles_v = validate_molecule(mol)
        if smiles_v:
            valid_smiles.append(smiles_v)
            valid_mols.append(Chem.MolFromSmiles(smiles_v))
        else:
            valid_mols.append(None)
            valid_smiles.append(None)
    ar = np.full((1, len(valid_mols)), True, dtype=bool)
    for index, v_smile in enumerate(valid_smiles):
        if v_smile not in smiles:
            ar[0][index] = v_smile
    return ar[0]


def novel_mean_score(mols, data):
    valid = novel_score(mols, data)
    return valid.mean()


def valid_mean(mols):
    valids = []
    for m in mols:
        try:
            smiles = Chem.MolToSmiles(m)
            valids.append(1)
        except Exception as e:
            valids.append(0)
    return np.array(valids).mean()


def valids(mols):
    # return x is not None and Chem.MolToSmiles(x) != ''
    valids = []
    try:
        for m in mols:
            try:
                smiles = Chem.MolToSmiles(m)
                valids.append(m)
            except Exception as e:
                valids.append(None)
    except TypeError as e:
        try:
            smiles = Chem.MolToSmiles(mols)
            valids.append(mols)
        except Exception as e:
            valids.append(None)
    return valids


def compute_SAS(mol):
    if mol is None:
        return 0
    fp = Chem.rdMolDescriptors.GetMorganFingerprint(mol, 2)
    fps = fp.GetNonzeroElements()
    score1 = 0.
    nf = 0
    # for bitId, v in fps.items():
    for bitId, v in fps.items():
        nf += v
        sfp = bitId
        score1 += SA_model.get(sfp, -4) * v
    try:
        score1 /= nf
    except ZeroDivisionError as e:
        return 0

    # features score
    nAtoms = mol.GetNumAtoms()
    nChiralCenters = len(Chem.FindMolChiralCenters(
        mol, includeUnassigned=True))
    ri = mol.GetRingInfo()
    nSpiro = Chem.rdMolDescriptors.CalcNumSpiroAtoms(mol)
    nBridgeheads = Chem.rdMolDescriptors.CalcNumBridgeheadAtoms(mol)
    nMacrocycles = 0
    for x in ri.AtomRings():
        if len(x) > 8:
            nMacrocycles += 1

    sizePenalty = nAtoms ** 1.005 - nAtoms
    stereoPenalty = math.log10(nChiralCenters + 1)
    spiroPenalty = math.log10(nSpiro + 1)
    bridgePenalty = math.log10(nBridgeheads + 1)
    macrocyclePenalty = 0.

    # ---------------------------------------
    # This differs from the paper, which defines:
    #  macrocyclePenalty = math.log10(nMacrocycles+1)
    # This form generates better results when 2 or more macrocycles are present
    if nMacrocycles > 0:
        macrocyclePenalty = math.log10(2)

    score2 = 0. - sizePenalty - stereoPenalty - \
             spiroPenalty - bridgePenalty - macrocyclePenalty

    # correction for the fingerprint density
    # not in the original publication, added in version 1.1
    # to make highly symmetrical molecules easier to synthetise
    score3 = 0.
    if nAtoms > len(fps):
        score3 = math.log(float(nAtoms) / len(fps)) * .5

    sascore = score1 + score2 + score3

    # need to transform "raw" value into scale between 1 and 10
    min = -4.0
    max = 2.5
    sascore = 11. - (sascore - min + 1) / (max - min) * 9.
    # smooth the 10-end
    if sascore > 8.:
        sascore = 8. + math.log(sascore + 1. - 9.)
    if sascore > 10.:
        sascore = 10.0
    elif sascore < 1.:
        sascore = 1.0

    return sascore


def _remap(x, x_min, x_max):
    return (x - x_min) / (x_max - x_min)


def synthetic_accessibility_score_scores(mols, norm=False):
    scores = [compute_SAS(mol) if mol is not None else None for mol in mols]
    scores = np.array(list(map(lambda x: 0 if x is None else x, scores)))
    scores = np.clip(_remap(scores, 5, 1.5), 0.0, 1.0) if norm else scores
    return scores


def natural_products_score(mols, norm=False):
    # calculating the score
    scores = [sum(NP_model.get(bit, 0)
                  for bit in Chem.rdMolDescriptors.GetMorganFingerprint(mol,
                                                                        2).GetNonzeroElements()) / float(
        mol.GetNumAtoms()) if mol is not None else None
              for mol in mols]

    # preventing score explosion for exotic molecules
    scores = list(map(lambda score: score if score is None else (
        4 + math.log10(score - 4 + 1) if score > 4 else (
            -4 - math.log10(-4 - score + 1) if score < -4 else score)), scores))

    scores = np.array(list(map(lambda x: -4 if x is None else x, scores)))
    scores = np.clip(_remap(scores, -3, 1), 0.0, 1.0) if norm else scores

    return scores


def _avoid_sanitization_error(op):
    try:
        return op()
    except ValueError:
        return None


def quantitative_estimation_druglikeness_scores(mols, norm=True):
    return np.array(list(map(lambda x: 0 if x is None else x, [
        _avoid_sanitization_error(lambda: QED.qed(mol)) if mol is not None else None for mol in
        mols])))


def water_octanol_partition_coefficient_scores(mols, norm=False):
    scores = [_avoid_sanitization_error(lambda: Crippen.MolLogP(mol)) if mol is not None else None
              for mol in mols]
    scores = np.array(list(map(lambda x: -3 if x is None else x, scores)))
    scores = np.clip(_remap(scores, -2.12178879609, 6.0429063424), 0.0, 1.0) if norm else scores
    return scores


def _compute_diversity(mol, fps):
    ref_fps = Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048)
    dist = DataStructs.BulkTanimotoSimilarity(ref_fps, fps, returnDistance=True)
    score = np.mean(dist)
    return score


def diversity_scores(mols, data):
    rand_mols = np.random.choice(data, 100)
    fps = [Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048) for mol in rand_mols]

    scores = np.array(
        list(map(lambda x: _compute_diversity(x, fps) if x is not None else 0, mols)))
    scores = np.clip(_remap(scores, 0.9, 0.945), 0.0, 1.0)
    return scores


def unique_total_score(mols):
    v = valids(mols)
    s = set()
    for mol in v:
        try:
            smile = Chem.MolToSmiles(mol)
            s.add(smile)
        except Exception as e:
            continue
    # s = set(map(lambda x: Chem.MolToSmiles(x), v))
    return 0 if len(v) == 0 else len(s) / len(v)


def _constant_bump(x, x_low, x_high, decay=0.025):
    return np.select(condlist=[x <= x_low, x >= x_high],
                     choicelist=[np.exp(- (x - x_low) ** 2 / decay),
                                 np.exp(- (x - x_high) ** 2 / decay)],
                     default=np.ones_like(x))


def novel_scores(mols, data):
    smiles = []
    for mol in mols:
        try:
            sm = Chem.MolToSmiles(mol)
            smiles.append(sm)
        except Exception as e:
            smiles.append(None)
    novel_l = []
    for smile in smiles:
        if smile not in data and smile is not None:
            novel_l.append(True)
        else:
            novel_l.append(False)
    novels = np.array(novel_l)
    return novels


def drugcandidate_scores(mols, data):
    scores = (_constant_bump(
        water_octanol_partition_coefficient_scores(mols, norm=True), 0.210, 0.945)
              + synthetic_accessibility_score_scores(mols, norm=True)
              + novel_scores(
                mols, data) + (1 - novel_scores(mols, data)) * 0.3) / 4

    return scores


def valid_scores(mols):
    t = map(valids, mols)
    l = list(t)
    trues = []
    for mo in l:
        if mo[0] is not None:
            trues.append(True)
        else:
            trues.append(False)
    score = np.array(trues, dtype=np.float32)
    return score


class MolecularMetrics(object):

    @staticmethod
    def _avoid_sanitization_error(op):
        try:
            return op()
        except ValueError:
            return None

    @staticmethod
    def remap(x, x_min, x_max):
        return (x - x_min) / (x_max - x_min)

    @staticmethod
    def valid_lambda(x):
        try:
            smiles = Chem.MolToSmiles(x)
        except Exception as e:
            x = None
        return x is not None and smiles != ''

    @staticmethod
    def valid_lambda_special(x):
        s = Chem.MolToSmiles(x) if x is not None else ''
        return x is not None and '*' not in s and '.' not in s and s != ''

    @staticmethod
    def valid_scores(mols):
        return np.array(list(map(MolecularMetrics.valid_lambda, mols)), dtype=np.float32)

    @staticmethod
    def valid_filter(mols):
        return list(filter(MolecularMetrics.valid_lambda, mols))

    @staticmethod
    def valid_total_score(mols):
        return np.array(list(map(MolecularMetrics.valid_lambda, mols)), dtype=np.float32).mean()

    @staticmethod
    def novel_scores(mols, smiles):
        return np.array(
            list(map(lambda x: MolecularMetrics.valid_lambda(x) and Chem.MolToSmiles(x) not in smiles, mols)))

    @staticmethod
    def novel_filter(mols, smiles):
        return list(filter(lambda x: MolecularMetrics.valid_lambda(x) and Chem.MolToSmiles(x) not in smiles, mols))

    @staticmethod
    def novel_total_score(mols, data):
        return MolecularMetrics.novel_scores(MolecularMetrics.valid_filter(mols), data).mean()

    @staticmethod
    def unique_scores(mols):
        smiles = list(map(lambda x: Chem.MolToSmiles(x) if MolecularMetrics.valid_lambda(x) else '', mols))
        return np.clip(
            0.75 + np.array(list(map(lambda x: 1 / smiles.count(x) if x != '' else 0, smiles)), dtype=np.float32), 0, 1)

    @staticmethod
    def unique_total_score(mols):
        v = MolecularMetrics.valid_filter(mols)
        s = set(map(lambda x: Chem.MolToSmiles(x), v))
        return 0 if len(v) == 0 else len(s) / len(v)

    @staticmethod
    def natural_product_scores(mols, norm=False):

        # calculating the score
        scores = [sum(NP_model.get(bit, 0)
                      for bit in Chem.rdMolDescriptors.GetMorganFingerprint(mol,
                                                                            2).GetNonzeroElements()) / float(
            mol.GetNumAtoms()) if mol is not None else None
                  for mol in mols]

        # preventing score explosion for exotic molecules
        scores = list(map(lambda score: score if score is None else (
            4 + math.log10(score - 4 + 1) if score > 4 else (
                -4 - math.log10(-4 - score + 1) if score < -4 else score)), scores))

        scores = np.array(list(map(lambda x: -4 if x is None else x, scores)))
        scores = np.clip(MolecularMetrics.remap(scores, -3, 1), 0.0, 1.0) if norm else scores

        return scores

    @staticmethod
    def quantitative_estimation_druglikeness_scores(mols, norm=False):
        return np.array(list(map(lambda x: 0 if x is None else x, [
            MolecularMetrics._avoid_sanitization_error(lambda: QED.qed(mol)) if mol is not None else None for mol in
            mols])))

    @staticmethod
    def water_octanol_partition_coefficient_scores(mols, norm=False):
        scores = [MolecularMetrics._avoid_sanitization_error(lambda: Crippen.MolLogP(mol)) if mol is not None else None
                  for mol in mols]
        scores = np.array(list(map(lambda x: -3 if x is None else x, scores)))
        scores = np.clip(MolecularMetrics.remap(scores, -2.12178879609, 6.0429063424), 0.0, 1.0) if norm else scores

        return scores

    # @staticmethod
    # def diversity_scores(mols, data):
    #     rand_mols = np.random.choice(data, 100)
    #     fps = [Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048) for mol in rand_mols]
    #
    #     scores = np.array(
    #         list(map(lambda x: MolecularMetrics.__compute_diversity(x, fps) if x is not None else 0, mols)))
    #     scores = np.clip(MolecularMetrics.remap(scores, 0.9, 0.945), 0.0, 1.0)
    #
    #     return scores

    @staticmethod
    def synthetic_accessibility_score_scores(mols, norm=False):
        scores = [MolecularMetrics._compute_SAS(mol) if mol is not None else None for mol in mols]
        scores = np.array(list(map(lambda x: 10 if x is None else x, scores)))
        scores = np.clip(MolecularMetrics.remap(scores, 5, 1.5), 0.0, 1.0) if norm else scores

        return scores

    @staticmethod
    def __compute_diversity(mol, fps):
        ref_fps = Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048)
        dist = DataStructs.BulkTanimotoSimilarity(ref_fps, fps, returnDistance=True)
        score = np.mean(dist)
        return score

    @staticmethod
    def drugcandidate_scores(mols, data):

        scores = (MolecularMetrics.constant_bump(
            MolecularMetrics.water_octanol_partition_coefficient_scores(mols, norm=True), 0.210,
            0.945)
                  # + MolecularMetrics.synthetic_accessibility_score_scores(mols,
                  #                                                          norm=True)
                  + MolecularMetrics.novel_scores(
                    mols, data) + (1 - MolecularMetrics.novel_scores(mols, data)) * 0.3) / 4

        return scores

    @staticmethod
    def constant_bump(x, x_low, x_high, decay=0.025):
        return np.select(condlist=[x <= x_low, x >= x_high],
                         choicelist=[np.exp(- (x - x_low) ** 2 / decay),
                                     np.exp(- (x - x_high) ** 2 / decay)],
                         default=np.ones_like(x))



from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.core.structure import Structure, PeriodicSite
from typing import Union, Tuple
import numpy as np
import scipy.constants as const
import pandas as pd

ANG_TO_BOHR = const.value("Angstrom star") / const.value("Bohr radius")


class SineCoulombMatrix(MaterialFeaturizer):
  """
  Create a fixed vector of length 118, containing rwa, fractional
  element compostitions in a compound.

  References
  ----------
  Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
     https://doi.org/10.1038/s41598-018-35934-y

  Examples
  --------
  >>> import jaqpotpy as jt
  >>> from pymatgen.core.lattice import Lattice
  >>> from pymatgen.core.structure import Structure
  >>> lattice = Lattice.cubic(4.2)
  >>> struct = Structure(lattice, ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
  >>> featurizer = jt.descriptors.material.SineCoulombMatrix()
  >>> features = featurizer.featurize(struct)
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (118,)
  Note
  ----
  This class requires matminer and Pymatgen to be installed.
  `NaN` feature values are automatically converted to 0 by this featurizer.
  """

  def __init__(self, max_atoms: int = 100, diag_elements: bool = True, flatten: bool = False):
    """
    Create a Sine Coulomb Matrix for a pymatgen.core.Structure

    Parameters
    ----------
    max_atoms: int (default 100)
      Maximum number of atoms for any crystal in the dataset. Used to
      pad the Coulomb matrix.
    diag_elems (bool): flag indication whether (True, default) to use
      the original definition of the diagonal elements; if set to False,
      the diagonal elements are set to 0
    flatten: bool (default True)
      Return flattened vector of matrix eigenvalues.

      References
      ----------
      Faber et al. "Crystal Structure Representations for Machine
         Learning Models of Formation Energies", Inter. J. Quantum Chem.
         115, 16, 2015. https://arxiv.org/abs/1503.07406
    """
    self.diag_elements = diag_elements
    self.flatten = flatten
    self.max_atoms = max_atoms
    return

  @property
  def __name__(self):
    return 'SineCoulombMatrix'

  def __getitem__(self):
    return self


  def _featurize(self, datapoint: Union[str, Structure], **kwargs) -> np.ndarray:
    """
    Calculate sine coulomb matrix for a structure.

    Parameters
    ----------
    datapoint: str (path of a file) or pymatgen.core.Structure
      Structure or path of a file to parse into a Structure

    Returns
    -------
    feats: np.ndarray
      Vector containing the sine coulomb matrix of a structure.
    """

    if isinstance(datapoint, str):
        if datapoint.split('.')[-1] == 'extxyz':
            from ase.io import read
            xyz = read(datapoint)
            lat = np.array(xyz.get_cell())
            sym = xyz.get_chemical_symbols()
            pos = xyz.arrays['positions']
            struct = Structure(lat, sym, pos)
        else:
            struct = Structure.from_file(datapoint)
    else:
        struct = datapoint

    sites = struct.sites
    atomic_numbers = np.array([site.specie.Z for site in sites])
    sin_mat = np.zeros((len(sites), len(sites)))
    coords = np.array([site.frac_coords for site in sites])
    lattice = struct.lattice.matrix

    for i in range(len(sin_mat)):
        for j in range(len(sin_mat)):
            if i == j:
                if self.diag_elements:
                    sin_mat[i][i] = 0.5 * atomic_numbers[i] ** 2.4
            elif i < j:
                vec = coords[i] - coords[j]
                coord_vec = np.sin(np.pi * vec) ** 2
                trig_dist = np.linalg.norm((np.asmatrix(coord_vec) * lattice).A1) * ANG_TO_BOHR
                sin_mat[i][j] = atomic_numbers[i] * atomic_numbers[j] / trig_dist
            else:
                sin_mat[i][j] = sin_mat[j][i]

    if self.flatten:
        eigs, _ = np.linalg.eig([sin_mat])
        zeros = np.zeros(self.max_atoms)
        zeros[:len(eigs[0])] = eigs[0]
        features = zeros
    else:
        features = pad_array([sin_mat], self.max_atoms)

    features = np.asarray(features)
    return features



  def _featurize_dataframe(self, datapoint: Union[str, Structure], **kwargs) -> pd.DataFrame:
    """
    Calculate sine coulomb matrix for a structure.

    Parameters
    ----------
    datapoint: str (path of a file) or pymatgen.core.Structure
      Structure or path of a file to parse into a Structure

    Returns
    -------
    feats: pd.DataFrame
      Vector containing the sine coulomb matrix of a structure.
    """
    self.flatten = True
    data = self._featurize(datapoint)
    columns = ['Feat_{}'.format(str(i)) for i in range(self.max_atoms)]
    return pd.DataFrame([data], columns=columns, index=[0])

def pad_array(x: list,
              shape: Union[Tuple, int],
              fill: float = 0.0,
              both: bool = False) -> np.ndarray:
  """
  Pad an array with a fill value.
  Parameters
  ----------
  x: np.ndarray
    A numpy array.
  shape: Tuple or int
    Desired shape. If int, all dimensions are padded to that size.
  fill: float, optional (default 0.0)
    The padded value.
  both: bool, optional (default False)
    If True, split the padding on both sides of each axis. If False,
    padding is applied to the end of each axis.
  Returns
  -------
  np.ndarray
    A padded numpy array
  """
  x = np.asarray(x)
  if not isinstance(shape, tuple):
    shape = tuple(shape for _ in range(x.ndim))
  pad = []
  for i in range(x.ndim):
    diff = shape[i] - x.shape[i]
    assert diff >= 0
    if both:
      a, b = divmod(diff, 2)
      b += a
      pad.append((a, b))
    else:
      pad.append((0, diff))
  pad = tuple(pad)  # type: ignore
  x = np.pad(x, pad, mode='constant', constant_values=fill)
  return x


from typing import Callable, List, Optional, Any
import numpy as np
import warnings
# from deepchem.utils.typing import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer

RDKitMol = Any
RDKitAtom = Any
RDKitBond = Any

import copy


# warnings.filterwarnings("ignore", category=DeprecationWarning)


class MordredDescriptors(MolecularFeaturizer):
    """Mordred descriptors.
      This class computes a list of chemical descriptors using Mordred.
      Please see the details about all descriptors from [1]_, [2]_.
      Attributes
      ----------
      descriptors: List[str]
        List of Mordred descriptor names used in this class.
      References
      ----------
      .. [1] Moriwaki, Hirotomo, et al. "Mordred: a molecular descriptor calculator."
         Journal of cheminformatics 10.1 (2018): 4.
      .. [2] http://mordred-descriptor.github.io/documentation/master/descriptors.html
      Note
      ----
      This class requires Mordred to be installed.
      Examples
      --------
      >>> import jaqpotpy as jt
      >>> smiles = ['CC(=O)OC1=CC=CC=C1C(=O)O']
      >>> featurizer = jt.descriptors.MordredDescriptors(ignore_3D=True)
      >>> features = featurizer.featurize(smiles)
      >>> type(features[0])
      <class 'numpy.ndarray'>
      >>> features[0].shape
      (1613,)
      """

    @property
    def __name__(self):
        return 'MordredDescriptors'

    def __init__(self, ignore_3D: bool = True):
        """
        Parameters
        ----------
        ignore_3D: bool, optional (default True)
          Whether to use 3D information or not.
        """
        self.ignore_3D = ignore_3D
        self.calc: Optional[Callable] = None
        self.descriptors: Optional[List] = None

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: RDKitMol, convert_nan: bool = True,**kwargs) -> np.ndarray:
        """
        Calculate Mordred descriptors.
        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object
        Returns
        -------
        np.ndarray
          1D array of Mordred descriptors for `mol`.
          If ignore_3D is True, the length is 1613.
          If ignore_3D is False, the length is 1826.
        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing
                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")

        feature = self.calc(datapoint)
        # convert errors to zero
        if convert_nan:
            feature = [
            -1000.0 if self.is_missing(val) or isinstance(val, str) else val
            for val in feature
        ]
        return np.asarray(feature)

    def _get_column_names(self, **kwargs) -> list:
        """
    Return the column names
    """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing
                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")
        from rdkit import Chem
        smiles = 'CC(=O)OC1=CC=CC=C1C(=O)O'
        mol = Chem.MolFromSmiles(smiles)
        feature = self.calc(mol).fill_missing(0).asdict()
        names = [key for key in feature.keys()]

        return names

    def _featurize_dataframe(self, datapoint: RDKitMol, convert_nan: bool = True, **kwargs) -> np.ndarray:
        """
    Calculate Mordred descriptors.
    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit Mol object
    Returns
    -------
    np.ndarray
      1D array of Mordred descriptors for `mol`.
      If ignore_3D is True, the length is 1613.
      If ignore_3D is False, the length is 1826.
    """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing
                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")
        feature = self.calc(datapoint)
        # convert errors to zero
        if convert_nan:
            feature = [
            -1000.0 if self.is_missing(val) or isinstance(val, str) else val
            for val in feature
        ]
        return np.asarray(feature)



from typing import List
from jaqpotpy.parsers.base_classes import Parser
from jaqpotpy.entities.material_models import (
    Pdb, Atoms
)
import pandas as pd


def _clean_key(word) -> str:
    return ''.join(i for i in word if i.isalnum())


def _str_to_num(s, start, stop, change):
    """
    string: the whole string
    start: the starting character
    stop: the stiping character
    change: the changing function --> 0: float() and 1: int()
    """

    x = s[start:stop]

    if x.strip() != '':
        if change == 0:
            x = float(x)

        else:
            x = int(x)
    return x


class PdbParser(Parser):
    """
    Pdb Parser.
    This class parses a pdb file (all pdb files in a folder) into a Pdb structure.
    For more information on the structure see jaqpotpy.models.material_models.Pdb

    Attributes
    ----------
    files_: str | List[str]
        File names of the pdb files that were parsed.

    References
    ----------
    .. [1] http://www.wwpdb.org/documentation/file-format-content/format33/sect9.html

    Examples
    --------
    >>> import jaqpotpy as jt
    >>> pdb_file = './AgNP.pdb'
    >>> parser = jt.parsers.pdb_parser.pdb_parser.PdbParser()
    >>> pdb = parser.parse(pdb_file)
    >>> type(pdb)
    generator
    >>> parsed = next(pdb)
    >>> type(parsed)
    jaqpotpy.models.material_models.Pdb
    """

    @property
    def __name__(self):
        return 'PdbParser'

    def __getitem__(self):
        return self

    def _parse(self, path) -> Pdb:

        """
        Parse pdb files.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        Returns
        -------
        jaqpotpy.models.material_models.Pdb
          A Pdb object
        """
        self.files_.append(path)

        # Initialize variables
        curr_key = ""
        pdb_dict: Pdb = Pdb(meta={}, atoms=Atoms(elements=[], coordinates=[], extraInfo=[]))
        extra = []
        lines = 0

        # Open the file and read it in as a list of rows
        with open(path) as f:
            pdb = f.read().splitlines()

        # Iterate through the file
        for row in pdb:
            curr_list = row.split()
            if curr_list[0] == "ATOM":  # Then there are specific characteristics about the atoms and we pass them to the JSON

                if curr_key != "ATOM":
                    if lines == 1:
                        pdb_dict.meta[curr_key] = extra[0]
                    else:
                        pdb_dict.meta[curr_key] = extra
                    curr_key = "ATOM"

                pdb_dict.atoms.elements.append(row[76:78].strip())
                pdb_dict.atoms.coordinates.append([
                    _str_to_num(row, 30, 38, 0), _str_to_num(row, 38, 46, 0), _str_to_num(row, 46, 54, 0)
                ])
                if row[22:26].strip() == "":
                    a = 0
                else:
                    a = _str_to_num(row, 22, 26, 1)

                pdb_dict.atoms.extraInfo.append({
                    "serial": _str_to_num(row, 6, 11, 1),
                    "name": row[12:16].strip(),
                    "altLoc": row[16],
                    "resName": row[17:20].strip(),
                    "chainID": row[21],
                    "resSeq": a,
                    "iCode": row[26],
                    "occupancy": _str_to_num(row, 54, 60, 0),
                    "tempFactor": _str_to_num(row, 60, 66, 0),
                    "charge": row[78:].strip()
                })
            else:
                # In this case we are at the begining of the pdb and we collect the meta data.

                if _clean_key(curr_list[0]) == curr_key:  # Then it is the first loop
                    extra.append(row[len(curr_key):])
                    lines += 1
                else:
                    # We check if there are multiple rocords to this specific key, and if not then we pass tha value only.
                    if curr_key != "":
                        if lines == 1:
                            pdb_dict.meta[curr_key] = extra[0]
                        else:
                            pdb_dict.meta[curr_key] = extra

                    # We change the key and initialize the dictionary and the lines variable
                    curr_key = _clean_key(curr_list[0])
                    extra = []
                    extra.append(row[len(curr_key):])
                    lines = 1
        return pdb_dict

    def _parse_dataframe(self, file: Pdb, filename: str) -> pd.DataFrame:
        """
        Parse pdb files in a pandas dataframe.

        Parameters
        ----------
        file: jaqpotpy.models.material_models.Pdb
            The Pdb structrure of a parsed file

        filename: str
            The name of the parsed file

        Returns
        -------
        pd.DataFrame()
        """

        df = pd.DataFrame()
        for i in range(len(file.atoms.extraInfo)):
            d = file.atoms.extraInfo[i]
            d['file'] = filename
            d['element'] = file.atoms.elements[i]
            d['x'] = file.atoms.coordinates[i][0]
            d['y'] = file.atoms.coordinates[i][1]
            d['z'] = file.atoms.coordinates[i][2]

            df = pd.concat([df, pd.DataFrame(d, index=[0])]).reset_index(drop=True)

        return df



from typing import Dict, Callable


class Preprocesses:

    classes: Dict[str, Callable] = {}
    fitted_classes:  Dict[str, Callable] = {}

    classes_y: Dict[str, Callable] = {}
    fitted_classes_y:  Dict[str, Callable] = {}

    def __init__(self):
        pass

    @classmethod
    def register_preprocess_class(cls, class_name, class_):
        cls.classes[class_name] = class_

    @classmethod
    def register_fitted_class(cls, class_name, class_):
        cls.fitted_classes[class_name] = class_

    @classmethod
    def register_preprocess_class_y(cls, class_name, class_):
        cls.classes_y[class_name] = class_

    @classmethod
    def register_fitted_class_y(cls, class_name, class_):
        cls.fitted_classes_y[class_name] = class_

    def __getitem__(self):
        return self



from typing import Union, List, Optional, Any
import numpy as np
from scipy.spatial import Delaunay
from jaqpotpy.helpers.periodic_table.element import Element
from tqdm import tqdm
import warnings
import pandas as pd
from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from jaqpotpy.cfg import config
from jaqpotpy.entities.material_models import *
from jaqpotpy.parsers import *

class GeomDescriptors(MaterialFeaturizer):
    """Geometrical Descriptors.
      This class computes a list of of geometrical descriptors of materials based on the Delaunay tessellation and 14 physicochemical properties of the involved atoms.
        Namely:
            - Atomic Number
            - Atomic radius
            - Atomic Weight
            - Atomic Volume
            - Boiling Point
            - Density
            - Dipole Polarizability
            - Evaporation Heat
            - Fusion Heat
            - Radioactivity (0 for non radioactive atoms, 1 for radioactive atoms)
            - Lattice Constant
            - Thermal Conductivity
            - Specific Heat
            - Electronegativity (Pauling scale)

      Attributes
      ----------
      descriptors: Optional[list]
        List of descriptor names used in this class.
        Items of the list must be selected from the set:
            "atomic_number" | "atomic_radius" | "atomic_weight" | "atomic_volume" |
            "boiling_point" | "density" | "dipole_polarizability" | "evaporation_heat" |
            "fusion_heat" | "is_radioactive" | "lattice_constant" | "thermal_conductivity" |
            "specific_heat" | "en_pauling"

      Examples
      --------
      >>> import jaqpotpy as jt
      >>> pdb_file = './AgNP.pdb'
      >>> featurizer = jt.descriptors.material.GeomDescriptors()
      >>> features = featurizer.featurize(pdb_file)
      >>> type(features[0])
      <class 'numpy.ndarray'>
    """

    @property
    def __name__(self):
        return 'GeomDescriptors'

    def __init__(self, descriptors: List[str] = None):
        if descriptors:
            for d in descriptors:
                if d not in ["atomic_number", "atomic_radius", "atomic_weight", "atomic_volume", "boiling_point", "density", "dipole_polarizability",
                             "evaporation_heat", "fusion_heat", "is_radioactive", "lattice_constant", "thermal_conductivity", "specific_heat", "en_pauling"]:
                    raise ValueError('Descriptor {} is not supported from this Featurizer. Consider building your custom featurizer'.format(d))
            self.descriptors = descriptors
        else:
            self.descriptors = ["atomic_number", "atomic_radius", "atomic_weight", "atomic_volume", "boiling_point", "density", "dipole_polarizability",
                             "evaporation_heat", "fusion_heat", "is_radioactive", "lattice_constant", "thermal_conductivity", "specific_heat", "en_pauling"]

        self.data = pd.DataFrame()

    def __getitem__(self):
        return self

    def _featurize(self, material: Union[str, Pdb, Xyz, Sdf], **kwargs) -> np.ndarray:
        """
        Calculate Geometrical Descriptors.
          This class computes a list of of geometrical descriptors of materials based on the Delaunay tessellation and 14 physicochemical properties of the involved atoms.
            Namely:
                - Atomic Number
                - Atomic radius
                - Atomic Weight
                - Atomic Volume
                - Boiling Point
                - Density
                - Dipole Polarizability
                - Evaporation Heat
                - Fusion Heat
                - Radioactivity (0 for non radioactive atoms, 1 for radioactive atoms)
                - Lattice Constant
                - Thermal Conductivity
                - Specific Heat
                - Electronegativity (Pauling scale)
        """
        warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

        ret_dataframe = pd.DataFrame()

        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False

        if isinstance(material, str):
            file_ext = material.split('.')[-1].lower()
            material_list = []
            if file_ext in ['sdf', 'mol']:
                parser = MolParser(material, file_ext)
            elif file_ext in ['pdb']:
                parser = PdbParser(material, file_ext)
            elif file_ext in ['xyz', 'extxyz']:
                parser = XyzParser(material, file_ext)
            else:
                raise ValueError(f'Files with extention {file_ext} are not supported. Supported file formats are "xyz","extxyz","pdb","sdf" and "mol".')

            for mat in parser.parse():
                material_list.append(mat)
        else:
            material_list = [material]


        for i, item in enumerate(tqdm(material_list, desc='Parsing files', disable=disable_tq)):
            obj = item.get_atoms()

            points = obj.coordinates
            atoms = obj.elements

            # Creation of tetrahedra of neighboring atoms.
            delaunay_tess = Delaunay(points)
            tetrahedra = delaunay_tess.simplices

            # Initialization of local variables
            freq = {}
            new_line = pd.DataFrame(index=[0])

            # Loop through all tetrahedra
            for tetrahedron in tetrahedra:

                # Create the tetrahedron key (e.g. Tetrahedron for elements C,H,N,O -> descriptor = CHNO)
                lst = [atoms[i].upper().strip() for i in tetrahedron]
                descriptor = ''.join(sorted(lst))

                # If the tetrahedron has already been found, increase it's frequency.
                if descriptor in [col[:col.find('_')] for col in new_line.columns]:
                    freq[descriptor] += 1

                # If the tetrahedron has not been found, calculate the total physicochemical properties of the tetrahedron and store add them to the DataFrame.
                else:
                    # Initialization
                    an = 0  # Atomic Number
                    ar = 0  # Atomic Radius
                    aw = 0  # Atomic Weight
                    av = 0  # Atomic Volume
                    bp = 0  # Boiling Point
                    de = 0  # Density
                    dp = 0  # Dipole Polarizability
                    eh = 0  # Evaporation Heat
                    fh = 0  # Fusion Heat
                    ra = 0  # Radioactivity
                    lc = 0  # Lattice Constant
                    tc = 0  # Thermal Conductivity
                    sh = 0  # Specific Heat
                    en = 0  # Electronegativity (Pauling scale)

                    # Loop through all atoms of the tetrahedron and add the according properties.
                    for i in lst:
                        el = Element(correct_element_format(i))
                        elmnt = el.parameters
                        an += elmnt.atomic_number
                        ar += elmnt.atomic_radius
                        aw += elmnt.atomic_weight
                        av += elmnt.atomic_volume
                        bp += elmnt.boiling_point
                        de += elmnt.density
                        if elmnt.dipole_polarizability:
                            dp += elmnt.dipole_polarizability
                        if elmnt.evaporation_heat:
                            eh += elmnt.evaporation_heat
                        if elmnt.fusion_heat:
                            fh += elmnt.fusion_heat
                        if elmnt.is_radioactive:
                            ra += elmnt.is_radioactive.bit_length()
                        if elmnt.lattice_constant:
                            lc += elmnt.lattice_constant
                        if elmnt.thermal_conductivity:
                            tc += elmnt.thermal_conductivity
                        if elmnt.specific_heat:
                            sh += elmnt.specific_heat
                        if elmnt.en_pauling:
                            en += elmnt.en_pauling

                    guide = {
                        "atomic_number": an,
                        "atomic_radius": ar,
                        "atomic_weight": aw,
                        "atomic_volume": av,
                        "boiling_point": bp,
                        "density": de,
                        "dipole_polarizability": dp,
                        "evaporation_heat": eh,
                        "fusion_heat": fh,
                        "is_radioactive": ra,
                        "lattice_constant": lc,
                        "thermal_conductivity": tc,
                        "specific_heat": sh,
                        "en_pauling": en
                    }

                    # Store the total physicochemical properties to the DataFrame
                    for d in self.descriptors:
                        new_line[descriptor + '_' + d] = guide[d]

                    # Set the found counter equal to 1.
                    freq[descriptor] = 1

            # Get the weighted average of the total physicochemical properties according to the frequency of the tetrahedra.
            for key, value in freq.items():
                value /= len(tetrahedra)
                for i in [col for col in new_line.columns if col[:4] == key]:
                    new_line[i] *= value

            ret_dataframe = pd.concat([ret_dataframe, new_line]).reset_index(drop=True)

        self.data = ret_dataframe

        return np.array(self.data)

    def _get_column_names(self, **kwargs) -> list:
        """
        Return the column names
        """
        return list(self.data.columns)

    def _featurize_dataframe(self, material, **kwargs) -> pd.DataFrame:
        """
        Calculate Mordred descriptors.
        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object
        Returns
        -------
        np.ndarray
          1D array of Mordred descriptors for `mol`.
          If ignore_3D is True, the length is 1613.
          If ignore_3D is False, the length is 1826.
        """
        _ = self._featurize(material)
        return self.data


def correct_element_format(element):
    """
    Correction of an elements format.

    Examples:
        AU -> Au
        C  -> C
        PB -> Pb

    element : str
        The element's name.
    """
    return ''.join([element[i] if i == 0 else element[i].lower() for i in range(len(element))])




# from tornado.web import RequestHandler
#
#
# class BaseClient(RequestHandler):
#
#     def get(self, *args, **kwargs):
#         self.write("say something")
#
#     def set_default_headers(self, *args, **kwargs):
#         self.set_header("Access-Control-Allow-Origin", "*")
#         self.set_header("Access-Control-Allow-Headers", "x-requested-with")
#         self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")



"""
Tests for Docking
"""
import os
import platform
import unittest
import pytest
import logging
import numpy as np
import jaqpotpy as jp
from jaqpotpy.descriptors import ComplexFeaturizer
from jaqpotpy.models import Model
from jaqpotpy.docking.pose_generation import PoseGenerator

IS_WINDOWS = platform.system() == 'Windows'


class TestDockingUtils(unittest.TestCase):
  """
  Does sanity checks on pose generation.
  """

  def setUp(self):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    # self.protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    # self.protein_file = os.path.join(current_dir, "7zb6.pdb")
    # self.ligand_file = os.path.join(current_dir, "mulno.sdf")
    # self.ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

    self.protein_file = os.path.join(current_dir, "1a9m_pocket.pdb")
    self.ligand_file = os.path.join(current_dir, "1a9m_ligand.sdf")


  @pytest.mark.slow
  def test_docker_init(self):
    """Test that Docker can be initialized."""
    vpg = jp.docking.VinaPoseGenerator()
    jp.docking.Docker(vpg)

  @unittest.skip("test only for conda env")
  def test_docking_utils(self):
      from jaqpotpy.docking.utils import create_hydrated_pdbqt_pdb
      protein_pdbqt = create_hydrated_pdbqt_pdb(self.protein_file, "./tmp")
      print(protein_pdbqt)


  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_dock(self):
    """Test that Docker can dock."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=10,
                                 num_modes=10,
                                 out_dir="./tmp")

    print(docked_outputs)
    # Check only one output since num_modes==1
    assert len(list(docked_outputs)) == 10



from jaqpotpy.utils.types import RDKitMol


ZINC_SMILES_CHARSET = ["C", "N", "O", "F", "B", "I", "H", "S", "P", "Cl", "Br"]


def add_H(mol: RDKitMol):
    mol.add_atom(1)
    return mol


def remove_H(mol: RDKitMol):
    mol.remove_atom(1)
    return mol


def add_B(mol: RDKitMol):
    mol.add_atom(5)
    return mol


def remove_B(mol: RDKitMol):
    mol.remove_atom(5)
    return mol


def add_I(mol: RDKitMol):
    mol.add_atom(53)
    return mol


def remove_I(mol: RDKitMol):
    mol.remove_atom(53)
    return mol


def add_C(mol: RDKitMol):
    mol.add_atom(6)
    return mol


def remove_C(mol: RDKitMol):
    mol.remove_atom(6)
    return mol


def add_N(mol: RDKitMol):
    mol.add_atom(7)
    return mol


def remove_N(mol: RDKitMol):
    mol.remove_atom(7)
    return mol


def add_O(mol: RDKitMol):
    mol.add_atom(8)
    return mol


def remove_O(mol: RDKitMol):
    mol.remove_atom(8)
    return mol


def add_F(mol: RDKitMol):
    mol.add_atom(9)
    return mol


def remove_F(mol: RDKitMol):
    mol.remove_atom(9)
    return mol


def add_P(mol: RDKitMol):
    mol.add_atom(15)
    return mol


def remove_P(mol: RDKitMol):
    mol.remove_atom(15)
    return mol


def add_S(mol: RDKitMol):
    mol.add_atom(16)
    return mol


def remove_S(mol: RDKitMol):
    mol.remove_atom(16)
    return mol


def add_Cl(mol: RDKitMol):
    mol.add_atom(17)
    return mol


def remove_Cl(mol: RDKitMol):
    mol.remove_atom(17)
    return mol


def add_Br(mol: RDKitMol):
    mol.add_atom(35)
    return mol


def remove_Br(mol: RDKitMol):
    mol.remove_atom(35)
    return mol



from pydantic import BaseModel
from typing import List, Optional


class Atoms(BaseModel):
    elements: List[str]
    coordinates: List[List[float]]
    extraInfo: Optional[list]


class Pdb(BaseModel):
    meta: dict
    atoms: Atoms

    def get_atoms(self):
        return self.atoms

class Xyz(BaseModel):
    num_atoms: int
    comment: str
    atoms: Atoms

    def get_atoms(self):
        return self.atoms


class SdfCountLine(BaseModel):
    n_atoms: int
    n_bonds: int
    n_atom_list: int
    chiral_flag: int
    n_stext_ents: int
    n_add_props: int
    version: str


class SdfBond(BaseModel):
    first_atom: int
    second_atom: int
    bond_type: int
    bond_stereo: int
    bond_topology: int
    reacting_center_status: int


class SdfHeader(BaseModel):
    title: str
    timestamp: str
    comment: str


class SdfCTable(BaseModel):
    count_line: SdfCountLine
    atoms: Atoms
    bonds: List[SdfBond]
    properties: list


class Sdf(BaseModel):
    header: SdfHeader
    con_table: SdfCTable
    extraInfo: Optional[list]

    def get_atoms(self):
        return self.con_table.atoms





from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.core.composition import Composition
from pymatgen.util.typing import CompositionLike
from typing import Any
from jaqpotpy.helpers.statistics import PropertyStats
import numpy as np
import pandas as pd


class ElementPropertyFingerprint(MaterialFeaturizer):
  """
  Fingerprint of elemental properties from composition.
  Based on the data source chosen, returns properties and statistics
  (min, max, range, mean, standard deviation, mode) for a compound
  based on elemental stoichiometry. E.g., the average electronegativity
  of atoms in a crystal structure. The chemical fingerprint is a
  vector of these statistics.
  See references for more details.
  References
  ----------
  .. [1] Pymatgen: Ong, S.P. et al. Comput. Mater. Sci. 68, 314-319 (2013).
  Examples
  --------
  >>> import jaqpotpy as jt
  >>> comp = "Fe2O3"
  >>> featurizer = jt.descriptors.material.ElementPropertyFingerprint()
  >>> features = featurizer.featurize([comp])
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (65,)
  Note
  ----
  This class requires matminer and Pymatgen to be installed.
  `NaN` feature values are automatically converted to 0 by this featurizer.
  """

  def __init__(self):
    """
    """
    self.ep_featurizer: Any = None
    self.pstats = PropertyStats()

  @property
  def __name__(self):
    return 'ElementPropertyFingerprint'

  def __getitem__(self):
    return self

  def _get_elemental_property(self, elem, property_name):
      if property_name == "block":
          block_key = {"s": 1.0, "p": 2.0, "d": 3.0, "f": 3.0}
          return block_key[getattr(elem, property_name)]
      else:
          value = getattr(elem, property_name)
          return np.nan if value is None else value

  def _featurize(self, datapoint: CompositionLike, **kwargs) -> np.ndarray:
    """
    Calculate chemical fingerprint from crystal composition.
    Parameters
    ----------
    datapoint: pymatgen.core.Composition object
      Composition object.
    Returns
    -------
    feats: np.ndarray
      Vector of properties and statistics derived from chemical
      stoichiometry. Some values may be NaN.
    """

    stats = ["minimum", "maximum", "range", "mean", "std_dev"]
    features = [
        "X",
        "row",
        "group",
        "block",
        "atomic_mass",
        "atomic_radius",
        "mendeleev_no",
        "electrical_resistivity",
        "velocity_of_sound",
        "thermal_conductivity",
        "melting_point",
        "bulk_modulus",
        "coefficient_of_linear_thermal_expansion",
    ]

    if not isinstance(datapoint, Composition):
        data = Composition(datapoint)
    else:
        data = datapoint

    all_attributes = []

    # Get the element names and fractions
    elements, fractions = zip(*data.element_composition.items())

    for attr in features:
      try:
        elem_data = [self._get_elemental_property(e, attr) for e in elements]
      except:
        raise ValueError('Cannot find property of Element.')

      for stat in stats:
        all_attributes.append(self.pstats.calc_stat(elem_data, stat, fractions))

    return np.array(all_attributes)

  def _featurize_dataframe(self, datapoint: CompositionLike, **kwargs) -> pd.DataFrame:
    """
    Calculate chemical fingerprint from crystal composition.
    Parameters
    ----------
    datapoint: pymatgen.core.Composition object
      Composition object.
    Returns
    -------
    feats: np.ndarray
      Vector of properties and statistics derived from chemical
      stoichiometry. Some values may be NaN.
    """

    stats = ["minimum", "maximum", "range", "mean", "std_dev"]
    features = [
        "X",
        "row",
        "group",
        "block",
        "atomic_mass",
        "atomic_radius",
        "mendeleev_no",
        "electrical_resistivity",
        "velocity_of_sound",
        "thermal_conductivity",
        "melting_point",
        "bulk_modulus",
        "coefficient_of_linear_thermal_expansion",
    ]

    if not isinstance(datapoint, Composition):
        data = Composition(datapoint)
    else:
        data = datapoint

    all_attributes = []

    # Get the element names and fractions
    elements, fractions = zip(*data.element_composition.items())

    for attr in features:
      try:
        elem_data = [self._get_elemental_property(e, attr) for e in elements]
      except:
        raise ValueError('Cannot find property of Element.')

      for stat in stats:
        all_attributes.append(self.pstats.calc_stat(elem_data, stat, fractions))

    columns = [i + '_' + j for j in features for i in stats]
    return pd.DataFrame(index=[0],columns=columns, data=np.array(all_attributes).reshape(1,-1))

